# Licensed under a 3-clause BSD style license - see LICENSE.rst
from dkist_processing_pac._version import __version__

from dkist_processing_pac import Data, CUModel, TelescopeModel, FittingFramework, FitCUParams, GenerateDemodMatrices, \
     FitM36Params, generic




import os
import pkg_resources
import yaml
import asdf
from configparser import ConfigParser
import numpy as np
try:
    from sunpy.coordinates import sun
except ImportError:
    raise ImportError('PAC Modules require sunpy>=1.0')
from dkist_processing_pac.data.required_params import elliptical
from dkist_processing_pac.tag import tag

#TODO: Rewrite based on Dave's new compute_telescope_geom to remove sunpy dependency
def compute_parallactic_angle(time):
    """ Calculate the parallactic angle of the Solar disk center at a specific time.

    All angles are in radians.

    Parameters
    ----------
    time : astropy.time.Time
        The absolute date/time at which to compute the parallactic angle

    Returns
    -------
    float
        The parallactic angle of disk center at the time specified [radians]

    """
    dkist_lat = 20.7047 * np.pi / 180
    dkist_long = -156.25 * np.pi / 180
    dec = sun.true_declination(time).value * np.pi / 180.
    ra = sun.true_rightascension(time).value * 15. * np.pi / 180.
    ha = time.sidereal_time('apparent', longitude=dkist_long * 180 / np.pi).value * 15 * np.pi / 180. - ra

    p = np.arctan2(np.sin(ha), np.tan(dkist_lat) * np.cos(dec) - np.sin(dec) * np.cos(ha))

    return p


#TODO: Update these defaults!
def gen_default_config(filename):
    """Generate an example fit configuration file with default options

    Parameters
    ----------
    filename : str
       Location to save example config file
    """
    config = ConfigParser(allow_no_value=True)

    config['Main'] = {'telescope_db': 'telescope_db.txt',
                      'method': 'differential_evolution'}

    with open(filename, 'w') as f:
        config.write(f)

    return

def init_fit_mode(fit_mode: str, wavelength: float, fit_TM: bool, init_set: str = 'default') -> dict:
    """Load a PA&C fit recipe from the database

    The recipe is returned as a ConfigParser object, which is basically a fancy dictionary

    Parameters
    ----------
    fit_mode : str
        Name of PA&C fit recipe

    Returns
    -------
    configparser.ConfigParser
        Object containing parameters, their ranges, and whether or not to vary them during a fit
    """
    mode_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/fit_modes/{}.yml'.format(fit_mode))
    if not os.path.exists(mode_file):
        raise FileNotFoundError("Could not find file for recipe '{}'".format(fit_mode))

    with open(mode_file, 'rb') as f:
        recipe = yaml.load(f, Loader=yaml.SafeLoader)
        print('{}: using PA&C fitting mode "{}"'.format(tag(), fit_mode))

    switches = recipe['switches']
    if not switches['fit_bounds']:
        print('{}: Boundless fit detected'.format(tag()))

    polcal_init_file = pkg_resources.resource_filename('dkist_processing_pac',
                                                       'data/init_values/polcal_{}.asdf'.format(init_set))
    if not os.path.exists(polcal_init_file):
        raise FileNotFoundError("Could not find the polcal init value file {}".format(polcal_init_file))
    with asdf.open(polcal_init_file, 'rb', lazy_load=False, copy_arrays=True) as f:
        polcal_init = f.tree
        print('{}: using initial values from set "{}"'.format(tag(), init_set))

    wave_idx = np.argmin(np.abs(polcal_init['wave'] - wavelength))

    mode_opts = {}
    for p in polcal_init['params'].keys():
        minv, val, maxv = polcal_init['params'][p][wave_idx, :]
        if 'dret' in p:
            vary = True
            if not switches['delta_d']:
                val = 0
                vary = False
        else:
            vary = recipe['vary'][p]

        if not switches['fit_bounds']:
            minv = -np.inf
            maxv = np.inf

        mode_opts[p] = {'min': minv,
                        'value': val,
                        'max': maxv,
                        'vary': vary}

    groupcal_init_file = pkg_resources.resource_filename('dkist_processing_pac',
                                                         'data/init_values/groupcal_{}.asdf'.format(init_set))
    if not os.path.exists(polcal_init_file):
        raise FileNotFoundError("Could not find the polcal init value file {}".format(polcal_init_file))
    with asdf.open(groupcal_init_file, 'rb', lazy_load=False, copy_arrays=True) as f:
        groupcal_init = f.tree

    wave_idx = np.argmin(np.abs(polcal_init['wave'] - wavelength))

    for p in groupcal_init['params'].keys():
        minv, val, maxv = groupcal_init['params'][p][wave_idx, :]
        if not switches['fit_bounds']:
            minv = -np.inf
            maxv = np.inf

        mode_opts[p] = {'min': minv,
                        'value': val,
                        'max': maxv,
                        'vary': fit_TM}

    mode_opts['switches'] = switches

    return mode_opts


def verify_fit_mode(mode_opts, fit_TM):
    """Make sure the loaded fit mode parameters are sufficient to actually run the fits

    All required model parameters need to exist and have an initial value.

    Parameters
    ----------
    mode_opts : configparser.ConfigParser
        Object containing fit parameter definitions. Each parameter in the model is a separate section with keywords
        'value', 'min', 'max', 'vary'.
    """

    missing_pars = []
    missing_vals = []
    bad_ranges = []
    for rp in elliptical['params']:
        if rp not in mode_opts.keys():
            missing_pars.append(rp)
        else:
            if rp != 'I_sys' and 'value' not in mode_opts[rp].keys():
                missing_vals.append(rp)
            if not mode_opts['switches']['fit_bounds'] \
                and (mode_opts[rp]['min'] != -np.inf or mode_opts[rp]['max'] != np.inf):
                bad_ranges.append(rp)

    if fit_TM:
        for p in ['x34', 't34', 'x56', 't56']:
            if p not in mode_opts.keys():
                missing_pars.append(p)
            else:
                if p != 'I_sys' and 'value' not in mode_opts[p].keys():
                    missing_vals.append(p)
                if not mode_opts['switches']['fit_bounds'] \
                    and (mode_opts[p]['min'] != -np.inf or mode_opts[p]['max'] != np.inf):
                    bad_ranges.append(p)

    if len(missing_pars) > 0:
        print('{}: ERROR: missing required fit parameters {}'.format(tag(), missing_pars))

    if len(missing_vals) > 0:
        print('{}: ERROR: missing initial values for parameters {}'.format(tag(), missing_vals))

    if len(bad_ranges) > 0:
        print('{}: ERROR: boundless fit requested, but the following parameters are bounded: {}'.format(tag(), bad_ranges))

    if len(missing_vals) > 0 or len(missing_pars) > 0 or len(bad_ranges) > 0:
        raise ValueError('Loaded fit parameters are not complete. See last status message(s) for info.')

def get_default_telescope_db():

    return pkg_resources.resource_filename('dkist_processing_pac', 'data/telescope_db.txt')


import numpy as np
from astropy.time import Time

from dkist_processing_pac.Data import Drawer
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from typing import List, Dict

from dkist_processing_pac.data import CONSTANTS
from dkist_processing_pac.tag import tag


class DCDrawer(Drawer):

    def __init__(self, fits_access_dict, skip_darks=True, remove_I_trend=True):
        super().__init__(data_dir=None, suffix='', remove_I_trend=False)
        self.fits_access_dict = dict()
        self.load_from_dict_of_objects(fits_access_dict, skip_darks=skip_darks)
        if remove_I_trend:
            self.find_clears()
            self.fit_intensity_trend()
        self.data_list = self.fits_access_dict

    def load_from_dict_of_objects(self, raw_fits_access: Dict[int, List[L0FitsAccess]],
                                  skip_darks=True):
        """Load all images found in a given directory. Data are stored as a list of astropy.io.fits HDUList objects.

        In addition to loading the actual data, header values are also inspected to create vectors of the polarizer and
        retarder angles, telescope geometry, and observation times.

        skip_darks : bool
            If True (default) then don't load any dark steps from the CS
        """
        inst_set = set()
        nummod_set = set()
        wave_set = set()
        ip_start_set = set()
        ip_end_set = set()
        final_step_num = 0
        for cs_step in sorted(raw_fits_access.keys()):
            meta_obj = raw_fits_access[cs_step][0]
            if meta_obj.gos_level0_status == 'DarkShutter' and skip_darks:
                continue

            inst_set.add(meta_obj.instrument)
            wave_set.add(meta_obj.wavelength)
            ip_start_set.add(meta_obj.ip_start_time)
            ip_end_set.add(meta_obj.ip_end_time)

            self.azimuth = np.append(self.azimuth, meta_obj.azimuth)
            self.elevation = np.append(self.elevation, meta_obj.elevation)
            self.table_angle = np.append(self.table_angle, meta_obj.table_angle)

            self.theta_pol_steps = np.append(self.theta_pol_steps, float(meta_obj.gos_polarizer_angle))
            self.theta_ret_steps = np.append(self.theta_ret_steps, float(meta_obj.gos_retarder_angle))
            self.pol_in = np.append(self.pol_in, meta_obj.gos_polarizer_status not in ['undefined', 'clear', False])
            self.ret_in = np.append(self.ret_in, meta_obj.gos_retarder_status not in ['undefined', 'clear', False])
            self.dark_in = np.append(self.dark_in, meta_obj.gos_level0_status == 'DarkShutter')

            self.timeobs = np.append(self.timeobs, Time(meta_obj.time_obs).mjd)
            nummod_set.add(meta_obj.number_of_modulator_states)
            self.fits_access_dict[final_step_num] = raw_fits_access[cs_step]  # So the index is still correct after skipping darks
            final_step_num += 1

        if len(nummod_set) > 1:
            raise ValueError('Not all input files have the same number of modulator states')
        self.nummod = nummod_set.pop()

        if len(inst_set) > 1:
            raise ValueError('Data belong to more than one instrument')
        self.instrument = inst_set.pop()

        if len(wave_set) > 1:
            raise ValueError('Data have more than one wavelength')
        self.wavelength = wave_set.pop()

        if len(ip_start_set) > 1 or len(ip_end_set) > 1:
            raise ValueError("Data don't all have the same IP start/end times")
        self.date_bgn = Time(ip_start_set.pop(), format='fits').mjd
        self.date_end = Time(ip_end_set.pop(), format="fits").mjd

        # TODO: Make this match Data (probably by changing Data???)
        # noise_floor_dict = CONSTANTS['Fallback_noise_floors']
        self.RN = 0.0

        self.numsteps = len(self.fits_access_dict.keys())

    def __getitem__(self, item):
        """This builtin method provides the mechanism by which the object can be sliced to provided a single SoCC.

        The slice must have two elements (x, y, l). E.g.:

        >>> I = DRWR[x, y, l]

        Parameters
        ----------
        item : tuple
            The (x, y, l) position tuple. Don't worry, python's slicing syntax will take care of this for you.

        Returns
        -------
        numpy.ndarray
            A 2D array of shape (M, N) where M is the number of modulator states and N is the number of steps in the
            Calibration Sequence.

        """
        if type(item) is not tuple or len(item) != 3:
            raise IndexError('Drawer must be indexed by exactly three values')

        xpos, ypos, lpos = item

        if not np.issubdtype(type(xpos), np.integer) \
                or not np.issubdtype(type(ypos), np.integer) \
                or not np.issubdtype(type(lpos), np.integer):
            raise IndexError('Only integers are allowed as valid indices')

        result = np.zeros((self.nummod, self.numsteps), dtype=np.float64)

        for n in range(self.numsteps):
            obj_list = self.fits_access_dict[n]
            modnum_list = [h.modulator_state for h in self.fits_access_dict[n]]
            for m in range(self.nummod):
                idx = modnum_list.index(m + 1)
                result[m, n] = obj_list[idx].data[xpos, ypos, lpos] / self.norm_func(self.timeobs[n])

        return result

    @property
    def shape(self):
        """The shape of this objects 'data'.

        This is useful for those who will be using the slicing functionality and expect to be able to see the shape of
        these data.
        """
        data = self.fits_access_dict[0][0].data

        return data.shape

    def find_clears(self):
        """Identify which HDUList objects contain clear observations and populate self.clear_objs

        A clear frame is defined as one in which both the polarizer and retarder were out of the light path.
        """
        clear_objs = []
        clear_times = np.array([])
        for n in range(self.numsteps):
            if not self.pol_in[n] and not self.ret_in[n] and not self.dark_in[n]:
                clear_objs.append(self.fits_access_dict[n])
                clear_times = np.append(clear_times, self.timeobs[n])

        self.clear_objs = clear_objs
        self.clear_times = clear_times

    def fit_intensity_trend(self):
        """Use clear frames to fit and remove any global intensity trends.

        The flux in each clear is averaged over all modulation states and the set of all clears is used to fit a linear
        trend of flux vs time. This line is then divided from all data, including clears.

        Note that because the absolute offset (i.e., intercept) is also fit the overall intensity is normalized by
        something very close to the flux in the first clear measurement.
        """
        try:
            avg_clear_flux = np.zeros(len(self.clear_objs))
        except AttributeError:
            print('{}: WARNING: clear frames have not yet been identified. Use .find_clears() to do so'.format(tag()))
            return
        if len(self.clear_objs) == 0:
            print('{}: WARNING: this Drawer does not contain any clear measurements. No correction is possible.'.format(tag()))
            return

        # First, get the linear trend in clear intensity
        for n in range(len(self.clear_objs)):
            tmp = 0.
            for j in range(self.nummod):
                tmp += np.nanmean(self.clear_objs[n][j].data)
            avg_clear_flux[n] = tmp / self.nummod

        self.I_clear = np.mean(avg_clear_flux)
        print('{}: Average flux in clear measurements (I_clear): {:<10.3f}'.format(tag(), self.I_clear))
        if self.I_clear < CONSTANTS['min_clear_flux']:
            raise ValueError('Flux in Clear measurements is too low ({:.2f} < {:.2f})'.format(self.I_clear,
                                                                                              CONSTANTS['min_clear_flux']))
        fit = np.poly1d(np.polyfit(self.clear_times, avg_clear_flux, 1)) / self.I_clear
        self.norm_func = fit



from configparser import ConfigParser
from glob import glob
import ast
import os
from dkist_processing_pac import FittingFramework, generic
from dkist_processing_pac.tag import tag


def main(top_level_dir, CU_output, TM_output, overwrite=False, save_stats=False, suffix='.FITS',
         telescope_db='telescope_db.txt', numCS=None, update_db=False, **kwargs):
    """Launch a fit for Telescope Model M36 group parameters using data from multiple PolCal sequences (i.e., a Dresser)

    The resulting best-fit values for the M36 parameters are pushed to the PA&C Telescope database iff ``update_db`` is
    True, otherwise they are simply printed to stdout and saved as header entries in the Telescope Mueller matrix output
    file. That Mueller matrix contains two HDUs; the first contains the Mueller matrix for M3 up to but not including
    M7 and the second contains the Mueller matrix for M12. The usefulness of this file is questionable because it the
    Mueller matrices will only be valid for a single geometry of the Telescope (which is specified in the header).

    Fitting M36 parameters requires multiple PolCal sequences (Drawers). Each Drawer should live in its own directory
    within the provided top-level directory. Each Drawer directory is assumed to have the following format: Each of the
    N steps in the CS exists as a single file with M HDUs (one for each modulator state), each with shape (X, Y). The
    resulting CU parameter file will have a single HDU with a single array of shape (X, Y, Z), where Z is the number of
    parameters in the CU model.

    Different combinations of free/fixed parameters are specified in the PA&C fitting recipe database. A recipe is
    loaded via the ``fit_mode`` keyword.

    Parameters
    ----------
    top_level_dir : str
        Location of directory containing sub-directories for each Drawer to be used for the fit

    TM_output : str
        Location in which to save a Telescope Mueller matrix FITS file. See above for file format and limitation.

    numCS : int
        Number of Drawers to use. If numCS < the total number of Drawers found in ``top_level_dir`` then only the first
        ``numCS`` drawers will be used.

    update_db : bool
        If True then the PA&C Telescope Model database will be updated with the best-fit values from the fit

    overwrite : bool
        If True then ``CU_par_output`` will be overwritten if it already exists

    save_stats : bool
        If True then a directory called "fitaux" will be created along side ``CU_par_output``. This directory will
        contain pickle objects with auxiliary fit information for use by QA routines

    suffix : str
        Only files in ``data_dir`` with this suffix will be loaded

    telescope_db : str
        Location of the file containing Telescope Model parameters as a function of time and wavelength

    kwargs : dict
        Any additional arguments to pass to the `FittingFramework`

    """

    if os.path.exists(TM_output) and not overwrite:
        raise IOError('File {} already exists and overwrite is False'.format(TM_output))

    # These next two lines are why I love python
    dirlist = sorted([i for i in glob('{}/*'.format(top_level_dir)) if os.path.isdir(i)])
    dirlist = dirlist[:numCS]

    if save_stats:
        dirname = os.path.dirname(TM_output) or '.'
        pickle_path = '{}/fitaux'.format(dirname)
        os.makedirs(pickle_path, exist_ok=True)
    else:
        pickle_path = None

    CU_params, TM_params, best_TM = FittingFramework.run_fits(dirlist, suffix=suffix, telescope_db=telescope_db,
                                                              pickle_path=pickle_path, fit_TM = True, **kwargs)

    print('{}: saving best-fit CU parameters to {}'.format(tag(), CU_output))
    CU_params.writeto(CU_output, overwrite=overwrite)
    CU_params.cleanup()
    print('{}: saving best-fit TM parameters to {}'.format(tag(), TM_output))
    TM_params.writeto(TM_output, overwrite=overwrite)
    TM_params.cleanup()

    if update_db:
        best_TM.save_to_database(telescope_db)

    return

def command_line():
    import argparse
    import sys

    parser = argparse.ArgumentParser(description='Use a set of Calibration Sequences to compute the Telescope mirror'
                                     "parameters. Accuracy is improved by using more CS's, but only to a point."
                                     'The data directory should contain a series of Calibration Sequences, each in its '
                                     'own subdirectory. Each CS should contain a single FITS file for each step in '
                                     'the CS that has already been reduced by instrument reduction code.')
    parser.add_argument('top_dir', help='Top-level directory containing Calibration Sequence subdirectories', nargs='?')
    parser.add_argument('CU_output', help='Location of FITS file to save Calibration Unit Parameters', nargs='?',
                        default='CMP.fits')
    parser.add_argument('TM_output', help='Location of FITS file to save Telescope Model Parameters', nargs='?',
                        default='TMP.fits')
    parser.add_argument('-N', '--numCS', type=int,
                        help="Number of CS's to use for the fit. If unspecified then all CS's found will be used")
    parser.add_argument('-t', '--telescope-db', help='Database listing telescope parameters as function of obstime and '
                                                     'wavelength',
                        nargs=1, default=[])
    parser.add_argument('-u', '--update-db', action='store_true',
                        help='Update the Telescope Database with the new values')
    parser.add_argument('-f', '--force', action='store_true', help='Overwrite outputs')
    parser.add_argument('-c', '--config', help='Load program parameters from supplied configuration INI file',
                        metavar='CONFIG_FILE')
    parser.add_argument('-m', '--fit-mode', help='The fitting recipe to use', default='baseline')
    parser.add_argument('-R', '--init-set', help='Name of the initial value set to use', default='default',
                        metavar='SET_NAME')
    parser.add_argument('-s', '--suffix', help='File suffix to filter data. File mask is data_dir/*suffix',
                        default='.FITS')
    parser.add_argument('-S', '--save-fit', action='store_true', help='Save lmfit-style fitting statistics')
    parser.add_argument('-d', '--default', help='Generate a configuration file with the standard default values',
                        metavar='CONFIG_FILE')
    parser.add_argument('-D', '--use-darks', action='store_false', help='Include any DARK CS steps in the fit')
    args = parser.parse_args()

    if args.default is not None:
        generic.gen_default_config(args.default)
        sys.exit(0)

    kwargs = {'fit_mode': args.fit_mode,
              'skip_darks': args.use_darks,
              'init_set': args.init_set}
    if not args.telescope_db:
        kwargs['telescope_db'] = generic.get_default_telescope_db()
    else:
        kwargs['telescope_db'] = args.telescope_db[0]
    if args.config is not None:
        config = ConfigParser()
        config.optionxform = str
        config.read(args.config)
        kwargs.update(dict(config['Main']))
        for k in kwargs.keys():
            try:
                kwargs[k] = ast.literal_eval(kwargs[k])
            except:
                pass

    if args.top_dir is None:
        print('You must specify the location of the top-level data directory')
        sys.exit(1)

    if os.path.exists(args.top_dir):
        sys.exit(main(args.top_dir, args.CU_output, args.TM_output, suffix=args.suffix, overwrite=args.force,
                      numCS=args.numCS, update_db=args.update_db, save_stats=args.save_fit, **kwargs))
    else:
        print('Directory {} does not exist. Aborting.'.format(args.data_dir))
        sys.exit(1)



import ast
import os
from configparser import ConfigParser
from dkist_processing_pac import FittingFramework, generic, Data
from dkist_processing_pac.tag import tag


def main(data_dir, CU_par_output, overwrite=False, save_stats=False, suffix='.FITS',
         telescope_db='telescope_db.txt', **kwargs):
    """Launch a fit for CU parameters on a single Drawer (i.e., data from a single PolCal sequence)

    The result is a single file that contains the best-fit CU parameters (and all parameters that were fixed). This file
    can then be resampled by the Instrument Partners to whatever sampling they need for the calculation of the actual
    demodulation matrices.

    Input data representing the PolCal sequence are read from the provided directory and assumed to have the following
    format. Each of the N steps in the CS exists as a single file with M HDUs (one for each modulator state), each
    with shape (X, Y). The resulting CU parameter file will have a single HDU with a single array of shape (X, Y, Z),
    where Z is the number of parameters in the CU model.

    Different combinations of free/fixed parameters are specified in the PA&C fitting recipe database. A recipe is
    loaded via the ``fit_mode`` keyword.

    Parameters
    ----------
    data_dir : str
        Location of the directory containing reduced data from a single PolCal sequence (i.e., a Drawer)

    CU_par_output : str
        Location in which to save the CU parameter file

    overwrite : bool
        If True then ``CU_par_output`` will be overwritten if it already exists

    save_stats : bool
        If True then a directory called "fitaux" will be created along side ``CU_par_output``. This directory will
        contain pickle objects with auxiliary fit information for use by QA routines

    suffix : str
        Only files in ``data_dir`` with this suffix will be loaded

    telescope_db : str
        Location of the file containing Telescope Model parameters as a function of time and wavelength

    kwargs : dict
        Any additional arguments to pass to the `FittingFramework`

    """
    if os.path.exists(CU_par_output) and not overwrite:
        raise IOError('File {} already exists and overwrite is False'.format(CU_par_output))

    print('{}: starting CU parameter fitting using SoCC found in {}'.format(tag(), data_dir))

    if save_stats:
        dirname = os.path.dirname(CU_par_output) or '.'
        pickle_path = '{}/fitaux'.format(dirname)
        os.makedirs(pickle_path, exist_ok=True)
    else:
        pickle_path = None

    CU_params, TM_params, _ = FittingFramework.run_fits([data_dir], suffix=suffix, fit_TM=False,
                                                        telescope_db=telescope_db, pickle_path=pickle_path, **kwargs)
    print('{}: saving best-fit CU parameters to {}'.format(tag(), CU_par_output))
    DRWR = Data.Drawer(data_dir, suffix=suffix)
    CU_params.writeto(CU_par_output, header=DRWR.data_list[0][0].header, overwrite=overwrite)
    CU_params.cleanup()
    TM_params.cleanup()

    return


def command_line():
    import argparse
    import sys

    parser = argparse.ArgumentParser(description='Compute demodulation matrices from a single Calibration Sequence. '
                                     'The data directory specified must contain a separate FITS file for each step in '
                                     'the CS that has already been processed by the instrument reduction code.')
    parser.add_argument('data_dir', help='Directory containing a single Calibration Sequence', nargs='?')
    parser.add_argument('output', help='Location of FITS file to save demodulation matrices', nargs='?',
                        default='dmod.fits')
    parser.add_argument('-f', '--force', action='store_true', help='Overwrite outputs')
    parser.add_argument('-c', '--config', help='Load program parameters from supplied configuration INI file',
                        metavar='CONFIG_FILE')
    parser.add_argument('-t', '--telescope-db', help='Database listing telescope parameters as function of obstime and '
                                                     'wavelength. This option is ignored if telescope_db is defined in '
                                                     'a provided CONFIG_FILE (-c).',
                        nargs=1, default=[])
    parser.add_argument('-s', '--suffix', help='File suffix to filter data. File mask is data_dir/*suffix',
                        default='.FITS')
    parser.add_argument('-S', '--save-fit', action='store_true', help='Save lmfit-style fitting statistics')
    parser.add_argument('-m', '--fit-mode', help='The fitting recipe to use', default='baseline')
    parser.add_argument('-R', '--init-set', help='Name of the initial value set to use', default='default',
                        metavar='SET_NAME')
    parser.add_argument('-d', '--default', help='Generate a configuration file with the standard default values',
                        metavar='CONFIG_FILE')
    parser.add_argument('-D', '--use-darks', action='store_false', help='Include any DARK CS steps in the fit')
    args = parser.parse_args()

    if args.default is not None:
        generic.gen_default_config(args.default)
        sys.exit(0)

    kwargs = {'fit_mode': args.fit_mode,
              'skip_darks': args.use_darks,
              'init_set': args.init_set}
    if not args.telescope_db:
        kwargs['telescope_db'] = generic.get_default_telescope_db()
    else:
        kwargs['telescope_db'] = args.telescope_db[0]

    if args.config is not None:
        config = ConfigParser()
        config.optionxform = str
        config.read(args.config)
        kwargs.update(dict(config['Main']))
        for k in kwargs.keys():
            try:
                kwargs[k] = ast.literal_eval(kwargs[k])
            except:
                pass

    if args.data_dir is None:
        print('You must specify the location of the data directory')
        sys.exit(1)

    if os.path.exists(args.data_dir):
        sys.exit(main(args.data_dir, args.output, suffix=args.suffix, overwrite=args.force,
                      save_stats=args.save_fit, **kwargs))
    else:
        print('Directory {} does not exist. Aborting.'.format(args.data_dir))
        sys.exit(1)


# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`
__all__ = ['gen_fake_data', 'gen_init_val_files', 'prep_fit']


import inspect
import multiprocessing as mp
import os


def tag():
    """Designed to be called from within other functions and return their module:func signature"""
    calling = inspect.stack()[1]
    info = inspect.getframeinfo(calling[0])
    proc = '{} - '.format(mp.current_process().name)
    if proc == 'MainProcess - ':
        proc = ''
    return '{}{}:{}:{}'.format(proc, os.path.basename(calling[1]), calling[3], info.lineno)


def is_mp():
    proc = '{} - '.format(mp.current_process().name)
    return not proc == 'MainProcess - '



import os
import uuid
from glob import glob
from types import SimpleNamespace
from typing import Optional
import numpy as np
from astropy.io import fits as pyfits
from astropy.time import Time
from dkist_processing_pac.tag import tag
from dkist_processing_pac.data.required_params import elliptical
from dkist_processing_pac.data import S122, CONSTANTS, load_visp_keywords, load_cryo_keywords, load_dlnirsp_keywords

"""The following basic functions simply return Mueller matrices of the named optical element. Parameters of each 
element can be specified, where appropriate."""


def linear_retarder(t_ret, ret, b=0):
    # See "DKIST Polarization Calibration: Modeling of Calibration Unit elements" by C. Beck, April 2018
    b_factor = np.sqrt(1. - b**2 / 4.)

    return t_ret * np.array([[   1, b/2.,                       0,                      0],
                             [b/2.,    1,                       0,                      0],
                             [   0,    0,  b_factor * np.cos(ret), b_factor * np.sin(ret)],
                             [   0,    0, -b_factor * np.sin(ret), b_factor * np.cos(ret)]], dtype=np.float64)

def elliptical_retarder(t_ret, d_h, d_45, d_r):
    # See either "Mueller matrix of DKIST calibration unit retarders" or
    #  "DKIST Polarization Calibration: Modeling of Calibration Unit elements"
    # both April 2018 by C. Beck.

    d = np.sqrt(d_h**2 + d_45**2 + d_r**2)
    cosd = np.cos(d)
    sind = np.sin(d)

    m22 = (d_h**2 + cosd * (d_45**2 + d_r**2)) / d**2
    m23 = ((1 - cosd) * d_45 * d_h) / d**2 + sind * d_r / d
    m24 = ((1 - cosd) * d_h * d_r) / d**2 - sind * d_45 / d
    m32 = ((1 - cosd) * d_45 * d_h) / d**2 - sind * d_r / d
    m33 = (d_45**2 + cosd * (d_h**2 + d_r**2)) / d**2
    m34 = ((1 - cosd) * d_45 * d_r) / d**2 + sind * d_h / d
    m42 = ((1 - cosd) * d_h * d_r) / d**2 + sind * d_45 / d
    m43 = ((1 - cosd) * d_45 * d_r) / d**2 - sind * d_h / d
    m44 = (d_r**2 + cosd * (d_45**2 + d_h**2)) / d**2

    return t_ret * np.array([[1,   0,   0,   0],
                             [0, m22, m23, m24],
                             [0, m32, m33, m34],
                             [0, m42, m43, m44]], dtype=np.float64)

def polarizer(t_pol, py, px=1):
    # See "DKIST Polarization Calibration: Modeling of Calibration Unit elements" by C. Beck, April 2018
    p_factor = px**2 + py**2
    p_ratio = (px**2 - py**2) / p_factor
    UV_factor = 2 * px * py / p_factor
    return 0.5 * t_pol * p_factor * np.array([[      1, p_ratio,           0,           0],
                                              [p_ratio,       1,           0,           0],
                                              [      0,       0,   UV_factor,           0],
                                              [      0,       0,           0,   UV_factor]], dtype=np.float64)


def rotation(theta):
    return np.array([[1,                  0,                 0, 0],
                     [0,  np.cos(2 * theta), np.sin(2 * theta), 0],
                     [0, -np.sin(2 * theta), np.cos(2 * theta), 0],
                     [0,                  0,                 0, 1]], dtype=np.float64)


def mirror(x, tau):
    # See "Mueller matrix of a mirror or mirror group", C. Beck March 2018

    mirror = np.array([[1 + x**2, 1 - x**2,                    0,                   0],
                       [1 - x**2, 1 + x**2,                    0,                   0],
                       [0,               0,  2 * x * np.cos(tau), 2 * x * np.sin(tau)],
                       [0,               0, -2 * x * np.sin(tau), 2 * x * np.cos(tau)]], dtype=np.float64)
    return mirror / (1 + x**2)


class Drawer():
    """Drawer is used to aggregate a collection of images corresponding to a Polarization Calibration Sequence and
    distribute a Set of Calibration Curves (SoCC) from those images.

    Aggregation is achieved by simply loading all FITS files in a directory and reading the configuration of the
    Calibration Unit from header values. These values are formed into Calibration Sequence (CS) vectors that can be
    iterated over by e.g., the CUModel.

    Distribution of the actual data is handled by slicing into this class. Each slice will provide a SoCC for a given
    (x, y, l) location. The result will be a single array that is M x N where M is the number of modulator states and N
    is the number of steps in the CS.
    In addition to the detector data, this class generates a set of vectors that describe the configuration of the
    Telescope and Calibration Unit during each exposure in the aggregated data.

    Finally, multiple instances of Drawer can be concatenated simply with the + operator.

    Attributes
    (only those you might need)
    ----------
    instrument : str
        The name of the instrument used to take the data (upper case)

    nummod : int
        The total number of modulator states in each step

    numsteps : int
        The total number of steps in the CS

    shape : tuple
        The shape of a single data exposure

    theta_pol_steps : numpy.ndarray
        The polarizer angle at each step in the CS

    theta_ret_steps : numpy.ndarray
        The retarder angle at each step in the CS

    pol_in : numpy.ndarray
        Boolean array - was the polarizer in the beam at each step?

    ret_in : numpy.ndarray
        Boolean array - was the retarder in the beam at each step?

    timeobs : numpy.ndarray
        The DATE-OBS keyword for each step of the CS. Format is MJD

    date_bgn : float
        Time at the start of the CS. Format is MJD

    date_end : float
        Time at the end of the CS. Format is MJD

    wavelength : float
        Coarse wavelength band of observations (nm)

    azimuth : numpy.ndarray
        Telescope azimuth at each step in the CS

    elevation : numpy.ndarray
        Telescope elevation at each step in the CS

    table_angle : numpy.ndarray
        Coude table angle at each step in the CS

    I_clear : float
        Average of the means of any detected Clear CS steps. Useful for initializing I_sys fit value.

    Notes
    -----
    An example:

    # Load the data from 'visp_CS_dir'.
    >>> DRWR = Drawer('visp_CS_dir')

    # Look at the number of modulator states and number of CS steps
    >>> DRWR.nummod
    8

    >>> len(DRWR.theta_pol_steps)
    20

    # Grab the SoCC for the pixel at (10, 100, 0)
    >>> I = DRWR[10, 100, 0]
    >>> I.shape
    (8, 20)

    Data access is via astropy.io.fits and is therefore very memory-efficient.
    """

    def __init__(self, data_dir=None, suffix='.FITS', instrument=None, remove_I_trend=True, skip_darks=True):
        """Create a new class instance. Optionally, load data from a directory.

        Parameters
        ----------
        data_dir : str
            Location of a directory containing all data corresponding to a single Calibration Sequence

        suffix : str
             A filter string use to select specific files. The file mask is data_dir/*suffix

        instrument : str
            The "official" (ed: what the hell does that mean?) string of the instrument used to acquire the calibration
            data. Case insensitive.

        remove_I_trend : bool
            If True then any global intensity variations will be removed with a linear fit to any clear measurements. If
            the provided data set does not contain clear measurements this option has no effect.

        skip_darks : bool
            If True (default) then don't load any dark steps from the CS
        """
        if instrument is not None:
            self.instrument = instrument
            self.load_inst_keywords()
        else:
            self.instrument = ''
            self.camera = ''
            self.modid_key = None
            self.modnum_key = None
        self.data_list = []
        self.theta_pol_steps = np.array([])
        self.theta_ret_steps = np.array([])
        self.pol_in = np.array([], dtype=bool)
        self.ret_in = np.array([], dtype=bool)
        self.dark_in = np.array([], dtype=bool)
        self.timeobs = np.array([])
        self.nummod = 0
        self.numsteps = 0
        self.azimuth = np.array([])
        self.elevation = np.array([])
        self.table_angle = np.array([])
        self.date_bgn = np.inf
        self.date_end = -np.inf
        self.wavelength = 0.
        self.RN = 0.
        self.norm_func = np.poly1d([0., 1.])
        self.I_clear = 0.

        if data_dir is not None:
            self.load_from_dir(data_dir, suffix=suffix, skip_darks=skip_darks)

        if remove_I_trend:
            self.find_clears()
            self.fit_intensity_trend()

    def __add__(self, other):
        """Concatenate two Drawer objects.

        First, make sure the two object share relevant parameters and then just concatenate the rest of the attributes.

        Returns
        -------
        Drawer
            A new Drawer object
        """
        if self.instrument != other.instrument:
            raise ValueError('{} and {} appear to be from different instruments'.format(self, other))

        if self.nummod != other.nummod:
            raise ValueError('{} and {} do not have the same number of modulator states'.format(self, other))

        if self.wavelength != other.wavelength:
            raise ValueError('{} and {} were not taken at the same wavelength'.format(self, other))

        output = Drawer(instrument=self.instrument, remove_I_trend=False)

        if self.RN != other.RN:
            newRN = max([self.RN, other.RN])
            output.RN = newRN
            print('{}: Warning {} and {} do not have the same read noise. Setting to the highest ({})'.format(tag(),
                                                                                                              self,
                                                                                                              other,
                                                                                                              newRN))

        output.data_list = self.data_list + other.data_list
        output.theta_pol_steps = np.append(self.theta_pol_steps, other.theta_pol_steps)
        output.theta_ret_steps = np.append(self.theta_ret_steps, other.theta_ret_steps)
        output.pol_in = np.append(self.pol_in, other.pol_in)
        output.ret_in = np.append(self.ret_in, other.ret_in)
        output.dark_in = np.append(self.dark_in, other.dark_in)
        output.timeobs = np.append(self.timeobs, other.timeobs)
        output.azimuth = np.append(self.azimuth, other.azimuth)
        output.elevation = np.append(self.elevation, other.elevation)
        output.table_angle = np.append(self.table_angle, other.table_angle)
        output.date_bgn = min(self.date_bgn, other.date_bgn)
        output.date_end = max(self.date_end, other.date_end)
        output.numsteps = self.numsteps + other.numsteps
        output.nummod = self.nummod
        output.wavelength = self.wavelength

        return output

    def __repr__(self):

        return "<PolCal SoCC Drawer started at {} with (m,n) = ({}, {}) and shape = {}>".format(self.date_bgn,
                                                                                               self.nummod,
                                                                                               self.numsteps,
                                                                                               self.shape)

    @property
    def shape(self):
        """The shape of this objects 'data'.

        This is useful for those who will be using the slicing functionality and expect to be able to see the shape of
        these data.
        """
        data = self.data_list[0][1].data

        return data.shape

    def load_inst_keywords(self):

        if self.instrument.upper() == 'VISP':
            load_visp_keywords()
            self.modid_key = S122['ViSP_CurrentState']
            self.modnum_key = S122['ViSP_NumberOfModulatorStates']
        elif self.instrument.upper() == 'DLNIRSP':
            load_dlnirsp_keywords()
            self.modid_key = S122['DLN_CurrentState']
            self.modnum_key = S122['DLN_NumberOfModulatorStates']
        #TODO: Is this not CRYO-SP and CRYO-CI? I forget
        elif self.instrument.upper() == 'CRYO':
            load_cryo_keywords()
            self.modid_key = S122['CRSP_CurrentState']
            self.modnum_key = S122['CRSP_NumberOfModulatorStates']
        elif self.instrument.upper() == 'VTF':
            self.modid_key = S122['VTF_CurrentState']
            self.modnum_key = S122['VTF_NumberOfModulatorStates']
        else:
            raise ValueError('Could not find information for instrument {}'.format(self.instrument))

    def load_from_dir(self, data_dir, suffix='.FITS', skip_darks=True):
        """Load all images found in a given directory. Data are stored as a list of astropy.io.fits HDUList objects.

        In addition to loading the actual data, header values are also inspected to create vectors of the polarizer and
        retarder angles, telescope geometry, and observation times.

        Parameters
        ----------
        data_dir : str
            The location of the directory containing the data to load. Should contain all images related to a single CS

        suffix : str
            Only files that end in this suffix will be loaded from the directory.

        skip_darks : bool
            If True (default) then don't load any dark steps from the CS
        """
        filelist = sorted(glob('{}/*{}'.format(data_dir, suffix)))
        if len(filelist) == 0:
            raise FileNotFoundError('Could not find any files in {} with suffix {}'.format(data_dir, suffix))
        nummod_list = []

        ref_id = 0
        ref_hdu = pyfits.open(filelist[ref_id])
        while ref_hdu[0].header[S122['GOSLevel0']] in ['Dark', 'DarkShutter'] and skip_darks:
            ref_id += 1
            ref_hdu = pyfits.open(filelist[ref_id])

        if ref_hdu[1].header['NAXIS'] != 3:
            raise ValueError('Data do not appear to have 3 dimensions (NAXIS = {})'.format(ref_hdu[1].header['NAXIS']))
        ref_inst = ref_hdu[0].header[S122['Instrument']]

        # ref_cam = ref_hdu[0].header[S122['CameraName']]
        ref_start = ref_hdu[0].header['DATE-BGN']
        ref_end = ref_hdu[0].header['DATE-END']
        ref_wave = ref_hdu[0].header[S122['Wavelength']]
        ref_hdu.close()

        self.instrument = ref_inst
        # self.camera = ref_cam
        self.load_inst_keywords()

        RN_list = []
        numsteps = 0
        for f in filelist:
            hdus = pyfits.open(f)
            if hdus[0].header[S122['GOSLevel0']] in ['Dark', 'DarkShutter']:
                try:
                    RN_list.append(hdus[0].header[S122['ReadNoise']])
                except KeyError:
                    print('{}: Found no read noise information in DARK step {}'.format(tag(), f))
                if skip_darks:
                    print('{}: skipping DARK step {}'.format(tag(), f))
                    continue

            if hdus[1].header['NAXIS'] != 3:
                raise ValueError(
                        'Data do not appear to have 3 dimensions (NAXIS = {})'.format(ref_hdu[1].header['NAXIS']))

            if hdus[0].header[S122['Instrument']] != ref_inst:
                raise ValueError('Not all input files were taken with the same instrument')

            # if hdus[0].header[S122['CameraName']] != ref_cam:
            #     raise ValueError('Not all input files were taken with the same camera')

            # if hdus[0].header['DATE-BGN'] != ref_start or hdus[0].header['DATE-END'] != ref_end:
            #     raise ValueError('Not all input files have the same start/end dates')

            if hdus[0].header[S122['Wavelength']] != ref_wave:
                raise ValueError('Not all input files were taken at the same wavelength')

            self.azimuth = np.append(self.azimuth, hdus[0].header['TAZIMUTH'])
            self.elevation = np.append(self.elevation, hdus[0].header['TELEVATN'])
            self.table_angle = np.append(self.table_angle, hdus[0].header['TTBLANGL'])

            self.theta_pol_steps = np.append(self.theta_pol_steps, float(hdus[0].header[S122['PolarizerAngle']]))
            self.theta_ret_steps = np.append(self.theta_ret_steps, float(hdus[0].header[S122['RetarderAngle']]))
            self.pol_in = np.append(self.pol_in, hdus[0].header[S122['GOSPolarizer']] not in ['undefined', 'clear', False])
            self.ret_in = np.append(self.ret_in, hdus[0].header[S122['GOSRetarder']] not in ['undefined', 'clear', False])
            self.dark_in = np.append(self.dark_in, hdus[0].header[S122['GOSLevel0']] in ['Dark', 'DarkShutter'])

            try:
                time = Time(hdus[0].header['DATE-OBS'])
            except ValueError:
                raise ValueError('Could not parse date header date {} from {}'.format(hdus[0].header['DATE-OBS'], f))
            self.timeobs = np.append(self.timeobs, time.mjd)

            nummod_list.append(hdus[0].header[self.modnum_key])
            self.data_list.append(hdus)
            numsteps += 1

        if np.unique(nummod_list).size != 1:
            raise ValueError('Not all input files have the same number of modulator states')

        if len(RN_list) > 0:
            self.RN = np.mean(RN_list)
            print('{:}: Noise floor read to be {:.2f}'.format(tag(), self.RN))
        else:
            print('{}: Could not find Read Noise information for Drawer...'.format(tag()))
            self.RN = self.get_fallback_noise_floor(ref_hdu)

        self.nummod = nummod_list[0]
        self.numsteps += numsteps
        self.date_bgn = min(self.date_bgn, Time(ref_start, format='fits').mjd)
        self.date_end = max(self.date_end, Time(ref_end, format='fits').mjd)
        self.wavelength = ref_wave

    def find_clears(self):
        """Identify which HDUList objects contain clear observations and populate self.clears

        A clear frame is defined as one in which both the polarizer and retarder were out of the light path.
        """
        clears = []
        clear_times = np.array([])
        for i, h in enumerate(self.data_list):
            if not self.pol_in[i] and not self.ret_in[i] and not self.dark_in[i]:
                clears.append(h)
                clear_times = np.append(clear_times, self.timeobs[i])

        self.clears = clears
        self.clear_times = clear_times

    def fit_intensity_trend(self):
        """Use clear frames to fit and remove any global intensity trends.

        The flux in each clear is averaged over all modulation states and the set of all clears is used to fit a linear
        trend of flux vs time. This line is then divided from all data, including clears.

        Note that because the absolute offset (i.e., intercept) is also fit the overall intensity is normalized by
        something very close to the flux in the first clear measurement.
        """
        try:
            avg_clear_flux = np.zeros(len(self.clears))
        except AttributeError:
            print('{}: WARNING: clear frames have not yet been identified. Use .find_clears() to do so'.format(tag()))
            return
        if len(self.clears) == 0:
            print('{}: WARNING: this Drawer does not contain any clear measurements. No correction is possible.'.format(tag()))
            return

        # First, get the linear trend in clear intensity
        for i, h in enumerate(self.clears):
            tmp = 0.
            for j in range(self.nummod):
                tmp += np.nanmean(h[j + 1].data)
            avg_clear_flux[i] = tmp / self.nummod

        self.I_clear = np.mean(avg_clear_flux)
        print('{}: Average flux in clear measurements (I_clear): {:<10.3f}'.format(tag(), self.I_clear))
        if self.I_clear < CONSTANTS['min_clear_flux']:
            raise ValueError('Flux in Clear measurements is too low ({:.2f} < {:.2f})'.format(self.I_clear,
                                                                                              CONSTANTS['min_clear_flux']))
        fit = np.poly1d(np.polyfit(self.clear_times, avg_clear_flux, 1)) / self.I_clear
        self.norm_func = fit


    def __getitem__(self, item):
        """This builtin method provides the mechanism by which the object can be sliced to provided a single SoCC.

        The slice must have two elements (x, y, l). E.g.:

        >>> I = DRWR[x, y, l]

        Parameters
        ----------
        item : tuple
            The (x, y, l) position tuple. Don't worry, python's slicing syntax will take care of this for you.

        Returns
        -------
        numpy.ndarray
            A 2D array of shape (M, N) where M is the number of modulator states and N is the number of steps in the
            Calibration Sequence.

        """
        if type(item) is not tuple or len(item) != 3:
            raise IndexError('Drawer must be indexed by exactly three values')

        xpos, ypos, lpos = item

        if not np.issubdtype(type(xpos), np.integer) \
                or not np.issubdtype(type(ypos), np.integer) \
                or not np.issubdtype(type(lpos), np.integer):
            raise IndexError('Only integers are allowed as valid indices')

        result = np.zeros((self.nummod, self.numsteps), dtype=np.float64)

        for n in range(self.numsteps):
            hdus = self.data_list[n]
            modnums = np.array([h.header[self.modid_key] for h in hdus[1:]])
            for m in range(self.nummod):
                idx = np.where(modnums == m + 1)[0][0]
                result[m, n] = hdus[idx + 1].data[xpos, ypos, lpos] / self.norm_func(self.timeobs[n])
                del hdus[idx + 1].data

        return result

    def get_uncertainty(self, data: np.ndarray) -> np.ndarray:
        """Right now this just computes a very simply noise estimate. In the future it will be able to read from
        uncertainty frames provided by the IPAs"""

        return np.sqrt(np.abs(data) + self.RN**2)

    def get_fallback_noise_floor(self, hdu: pyfits.HDUList) -> float:
        """Used to provide a absolute last-resort noise floor if the instruments haven't provided their own
        """

        noise_floor_dict = CONSTANTS['Fallback_noise_floors']
        if self.instrument.upper() == 'VISP':
            noise_floor = noise_floor_dict['VISP']

        elif self.instrument.upper() == 'DLNIRSP':
            print('$$', self.camera)
            if self.camera == 'DLVIS':
                noise_floor = noise_floor_dict['DLNIRSP_VIS']
            elif self.camera == 'DLNIR':
                noise_floor = noise_floor_dict['DLNIRSP_NIR']
            else:
                raise ValueError('Could not find entry for DLNIRSP camera {}'.format(self.camera))

        elif self.instrument.upper() == 'CRYO':
            cryo_inst = hdu[0].header[S122['CRSP_CryoInstrument']]
            if cryo_inst == 1:
                noise_floor = noise_floor_dict['CRYO_SP']
            elif cryo_inst == 2:
                noise_floor = noise_floor_dict['CRYO_CI']
            else:
                raise ValueError('Could not find entry for Cryo instrument {}'.format(cryo_inst))

        elif self.instrument.upper() == 'VTF':
            noise_floor = noise_floor_dict['VTF']

        else:
            noise_floor = noise_floor_dict['Last_resort']
            print('{}: Could not figure out what noise floor to use for instrument "{}". '
                  'Falling back to default value.'.format(tag(), self.instrument))

        print('{:}: Using default noise floor of {:.2f}'.format(tag(), noise_floor))

        return noise_floor

    def plot_curves(self, output, location=None):
        """Plot Calibration Curves for all points in the data set

        Each (x, y, l) location will be given a separate plot containing overplots of all modulation states

        Parameters
        ----------
        output : str
            Location to save plot file
        """
        from cycler import cycler
        import matplotlib
        import matplotlib.pyplot as plt
        from matplotlib.backends.backend_pdf import PdfPages as PDF
        matplotlib.style.use('classic')

        matplotlib.rc('axes', prop_cycle=cycler(color=['#a6cee3','#1f78b4','#b2df8a','#33a02c','#fb9a99','#e31a1c',
                  '#fdbf6f','#ff7f00','#cab2d6','#6a3d9a','#ffff99','#b15928']))

        numx, numy, numl = self.shape
        if location is not None:
            x_range = [location[0]]
            y_range = [location[1]]
            l_range = [location[2]]
        else:
            x_range = range(numx)
            y_range = range(numy)
            l_range = range(numl)

        pp = PDF(output)
        for i in x_range:
            for j in y_range:
                for k in l_range:
                    ax = plt.figure().add_subplot(111)
                    ax.set_xlabel('CS step')
                    ax.set_ylabel('Counts')
                    d = self[i, j, k]
                    for m in range(self.nummod):
                        ax.plot(d[m])
                    ax.set_title('[{}, {}, {}]'.format(i, j, k))
                    pp.savefig(ax.figure)
                    del d

        pp.close()
        plt.close('all')

class Dresser():
    """Dresser is, quite simply, a collection of Drawer objects. It's primary function is to allow the
    `FittingFramework` to not have to worry about how many Drawers the user has provided for, e.g., fitting the M36
    group parameters.

    This class doesn't do much except keep track of the Drawers and concatenate a StoCCinG when a user asks for a
    specific location.

    Its attributes are basically the same as those of Drawer

    Attributes
    (only those you might need)
    ----------
    instrument : str
        The name of the instrument used to take the data (upper case)

    nummod : int
        The total number of modulator states in each step

    numsteps : int
        The total number of steps in all CS's across all Drawers

    shape : tuple
        The shape of a single data exposure

    theta_pol_steps : numpy.ndarray
        The polarizer angle at each CS step

    theta_ret_steps : numpy.ndarray
        The retarder angle at each CS step

    pol_in : numpy.ndarray
        Boolean array - was the polarizer in the beam at each step?

    ret_in : numpy.ndarray
        Boolean array - was the retarder in the beam at each step?

    date_bgn : float
        Time at the start of the CS of the first Drawer. Format is MJD

    date_end : float
        Time at the end of the CS of the last Drawer. Format is MJD

    wavelength : float
        Coarse wavelength band of observations (nm)

    azimuth : numpy.ndarray
        Telescope azimuth at each CS step

    elevation : numpy.ndarray
        Telescope elevation at each CS step

    table_angle : numpy.ndarray
        Coude table angle at each CS step

    I_clear : float
        Average of the means of any detected Clear CS steps. Useful for initializing I_sys fit value.
    """

    def __init__(self):
        """Create a new class instance. This is easy; just initialize all the arrays that might eventually hold stuff.
        """
        self.drawers = []
        self.azimuth = np.array([])
        self.elevation = np.array([])
        self.table_angle = np.array([])
        self.theta_pol_steps = np.array([])
        self.theta_ret_steps = np.array([])
        self.pol_in = np.array([], dtype=bool)
        self.ret_in = np.array([], dtype=bool)
        self.dark_in = np.array([], dtype=bool)
        self.delta_t = np.array([])
        self.nummod = 0
        self.numsteps = 0
        self.date_bgn = np.inf
        self.date_end = -np.inf
        self.wavelength = 0.
        self.instrument = None
        self.shape = ()

    def __repr__(self):

        rprstr = "<PolCal Dresser with (m,n) = ({}, {}) and the following Drawers:\n".format(self.nummod, self.numsteps)
        for d in self.drawers:
            rprstr += " {}\n".format(d)
        rprstr += ">"

        return rprstr

    @property
    def numdrawers(self):
        """The number of Drawers (AKA CS's) in this Dresser

        Returns
        -------
        int
        """
        return len(self.drawers)

    @property
    def drawer_step_list(self):
        """The number of CS steps in each Drawer

        Returns
        -------
        list
        """
        return [d.numsteps for d in self.drawers]

    def add_drawer(self, drawer: Drawer):
        """Add a Drawer to the Dresser

        Checks are made to ensure the Drawer fits in the Dresser (correct instrument, number of positions, etc.), and
        the CS configuration vectors are updated to include information from the new Drawer.

        Parameters
        ----------
        drawer : `Data.Drawer`
            PolCal Data from a single CS
        """
        if self.nummod == 0:
            self.nummod = drawer.nummod
        else:
            if drawer.nummod != self.nummod:
                raise ValueError('Trying to add Drawer with {} mod states to Dresser with {}'.format(drawer.nummod,
                                                                                                     self.nummod))

        if self.wavelength == 0.:
            self.wavelength = drawer.wavelength
        else:
            if drawer.wavelength != self.wavelength:
                raise ValueError('Drawer with wave = {:6.1f} cannot be added to Dresser with wave = {:6.1f}'.\
                                 format(drawer.wavelength, self.wavelength))

        if self.instrument is None:
            self.instrument = drawer.instrument
        else:
            if drawer.instrument != self.instrument:
                raise ValueError('Drawer from instrument {} cannot be added to Dresser from instrument {}'.\
                                 format(drawer.instrument, self.instrument))

        if self.shape == ():
            self.shape = drawer.shape
        else:
            if drawer.shape != self.shape:
                raise ValueError('Drawer with shape {} does not fit into Dresser with shape {}'.format(drawer.shape,
                                                                                                       self.shape))

        self.drawers.append(drawer)
        for attr in ['azimuth', 'elevation', 'table_angle', 'theta_pol_steps',
                     'theta_ret_steps', 'pol_in', 'ret_in', 'dark_in']:
            setattr(self, attr, np.append(getattr(self, attr), getattr(drawer, attr)))

        self.delta_t = np.append(self.delta_t, drawer.timeobs - np.min(drawer.timeobs))

        self.date_bgn = min(self.date_bgn, drawer.date_bgn)
        self.date_end = max(self.date_end, drawer.date_end)
        self.numsteps += drawer.numsteps


    @property
    def I_clear(self):
        """The I_clear property for each Drawer

        Returns
        -------
        `numpy.ndarray`
        """
        I_clear = np.empty((self.numdrawers,), dtype=np.float64)
        for i, d in enumerate(self.drawers):
            I_clear[i] = d.I_clear

        return I_clear

    def __getitem__(self, item):
        """Concatenate all the SoCCs at the same position from each Drawer.

        We call this a Set of Calibration Curves in a Group (a StoCCinG).

        Parameters
        ----------
        item: tuple
            The (x, y, l) position tuple. Don't worry, python's slicing syntax will take care of this for you.

        Returns
        -------
        numpy.ndarray
            Array of shape (M, N) where M is the number of modulator states and N is the total number of steps across
            all CS's in all Drawers.
        """
        data = np.empty((self.nummod, self.numsteps), dtype=np.float64)
        uncertainty = np.empty(data.shape, dtype=np.float64)

        xpos, ypos, lpos = item

        idx = 0
        for d in self.drawers:
            drawer_data = d[xpos, ypos, lpos]
            data[:, idx:idx + d.numsteps] = drawer_data
            uncertainty[:, idx:idx + d.numsteps] = d.get_uncertainty(drawer_data)
            idx += d.numsteps

        return data, uncertainty

class CUModelParams():
    """A container for CU Model parameters. The primary functions of this object are to write a FITS file that contains
    these parameters in a format that allows Instrument Partners to resample them to whatever field they need and to
    load a set of parameters after they have been resampled.

    The CU parameters are kept in a 3D array of shape (X, Y, L, Z) where (X, Y, L) is the shape of the FOV present in
    the Drawer(s) used in fitting and Z is the number of CU parameters (fixed and free).
    """

    def __init__(self,
                 x: int,
                 y: int,
                 l: int,
                 numdrawers: int,
                 nummod: int,
                 fit_mode: str = 'baseline',
                 init_set: str = 'default',
                 header: Optional[pyfits.header.Header] = None):
        """Initialize the object. All this does is create the empty array that will hold CU parameters.

        Parameters
        ----------
        x : int
            Number of field spatial positions in the x direction

        y : int
            Number of field spatial positions in the y direction

        l : int
            Number of field positions in the l dimension

        numdrawers : int
            Total number of Drawers that were used to fit CU parameters. This is greater than 1 when fitting Telescope
            M36 group parameters

        fit_mode : str
            Name of the PA&C fitting recipe used to produce the CU parameters

        init_set : str
            Name of the initial value set used to produce CU parameters

        header : `astropy.io.fits.header.Header`
            If provided, this object will be initialized with this header
        """
        self.shape = (x, y, l)
        self.numdrawers = numdrawers
        self.nummod = nummod

        numparams = numdrawers * (len(self.required['params']) - len(self.required['global'])) + len(self.required['global'])
        numparams += nummod * 4
        self.fit_mode = fit_mode
        self.init_set = init_set
        self.tmp_name = 'tmpCU_pars_{}.fits'.format(uuid.uuid4().hex)

        self._init_tmp_file(numparams, x, y, l, header=header)

        self.hdu_list = pyfits.open(self.tmp_name, mode='update')
        self.CU_params = self.hdu_list[0].data

    def __del__(self):
        """This is only here to be nice to any developers/testers who forget to call cleanup().
        In general, it is bad practice to rely on this method and cleanup() should be called explicitly.
        """
        try:
            self.cleanup()
        except:
            pass

    def _init_tmp_file(self, numparams, x, y, l, header=None):
        """Initialize a temp FITS file that will store the CU Parameters. We do this so that a a single file descriptor
        is used every time __setitem__ is called an the results are flushed to disk.

        Parameters
        ----------
        numparams : int
            The number of fit parameters

        x, y, l : int
            The shape of the Dresser

        header : `astropy.io.fits.header.Header`
            If provided, initialize the tmp file with this header
        """
        tmp_ph = pyfits.PrimaryHDU(np.zeros((x, y, l, numparams)), header=header)
        tmp_ph.header['PACFMODE'] = self.fit_mode
        tmp_ph.header['INITSET'] = self.init_set
        i = 0
        for p in self.required['global']:
            tmp_ph.header['CUPAR{:03n}'.format(i)] = (p, 'CU parameter index {}'.format(i))
            i += 1
        for j in range(self.numdrawers):
            for p in self.required['params']:
                if p in self.required['global']:
                    continue
                tmp_ph.header['CUPAR{:03n}'.format(i)] = ('{:}_CS{:02n}'.format(p, j),
                                                           'CU parameter index {} for CS {}'.format(i, j))
                i += 1
        for m in range(self.nummod):
            for s in ['I', 'Q', 'U', 'V']:
                tmp_ph.header['CUPAR{:03n}'.format(i)] = ('modmat_{}{}'.format(m,s),
                                                          'Modulation matrix entry {} -> {}'.format(m, s))
                i += 1

        tmp_ph.writeto(self.tmp_name)
        del tmp_ph.data
        del tmp_ph

    @property
    def required(self):
        return elliptical

    @property
    def par_names(self):
        """A list containing the names of the parameters that correspond to the last axis of self.CU_params"""
        names = self.required['global'][:]
        for i in range(self.numdrawers):
            names += ['{:}_{:02n}'.format(p, i) for p in self.required['params'] if p not in self.required['global']]

        for m in range(self.nummod):
            for s in ['I', 'Q', 'U', 'V']:
                names += ['modmat_{}{}'.format(m,s)]

        return names

    def __repr__(self):
        return "<CUModelParams with shape ({}, {}, {}, {}) from {} Drawers with {} fitting recipe>".\
            format(*list(self.CU_params.shape) + [self.numdrawers, self.fit_mode])

    def __setitem__(self, key, value):
        """Insert a set of fit parameters from a SoCC or StoCCinG into the CU parameter array.

        Every time an item is set, the current object will be written to disk. This is a somewhat janky way to ensure
        that a crash on looong fits does not result in a total loss of fit results.

        Parameters
        ----------
        key : tuple
            The (x, y, l) spatial location of the SoCC or StoCCinG

        value : `lmfit.MinimizerResult`
            The fitting object that is produced by lmfit's `minimize` function. It contains the best-fit CU parameters.
        """
        xpos, ypos, lpos = key

        if value is None:
            cu_pars = np.zeros(self.CU_params.shape[3])
            cu_pars[-self.nummod * 4::4] = 1. / self.nummod

        else:
            pardict = value.params.valuesdict()
            cu_pars = np.array([pardict[p] for p in self.required['global']])
            for i in range(self.numdrawers):

                cu_pars = np.append(cu_pars,
                                    np.array([pardict['{:}_CS{:02n}'.format(p,i)] \
                                              for p in self.required['params'] if p not in self.required['global']]))

            for m in range(self.nummod):
                for s in ['I', 'Q', 'U', 'V']:
                    cu_pars = np.append(cu_pars, pardict['modmat_{}{}'.format(m, s)])

        self.CU_params[xpos, ypos, lpos, :] = cu_pars
        self.hdu_list.flush()

    def __getitem__(self, item):
        """Use the currently loaded CU parameters to generate a dictionary of parameter : value pairs.

        This dictionary is intended for use in initializing a `CUModel.CalibrationSequence` object with the best-fit
        CU parameters.

        Parameters
        ----------
        item : tuple
            (x, y, l) location of the CU parameters to load

        Returns
        -------
        dict
            Dictionary of parameter : value pairs for all parameters in the CU model
        """
        xpos, ypos, lpos = item

        params = {}
        i = 0
        for p in self.required['global']:
            params[p] = self.CU_params[xpos, ypos, lpos, i]
            i += 1
        for j in range(self.numdrawers):
            for p in self.required['params']:
                if p in self.required['global']:
                    continue
                params['{:}_CS{:02n}'.format(p, j)] = self.CU_params[xpos, ypos, lpos, i]
                i += 1

        for m in range(self.nummod):
            for s in ['I', 'Q', 'U', 'V']:
                params['modmat_{}{}'.format(m, s)] = self.CU_params[xpos, ypos, lpos, i]
                i += 1

        return params

    def writeto(self, filename, header=None, overwrite=True):
        """Write the CU parameters to a FITS file

        The FITS file will have a single HDU with a (X, Y, L, Z). Any header provided will be updated to include the
        parameter names for each entry along the Z dimension. If no header is provided a blank one will be created.

        Parameters
        ----------
        filename : str
            Path to save the FITS file

        header : `astropy.io.fits.header.Header`
            An optional header to add to the FITS file. It will be updated with information about the parameter names

        overwrite : bool
            If False then an error will be raised if filename already exists
        """
        if header is not None:
            self.hdu_list[0].header.update(header)

        self.hdu_list.writeto(filename, overwrite=overwrite)

    def loadfits(self, filename):
        """Read CU parameters from a FITS file

        The FITS file should have a single HDU with an (X, Y, L, Z) array. The header MUST contain the header keywords
        generated by `writeto`; these are compared with the parameter list to verify the format/usefulness of the file.

        Parameters
        ----------
        filename : str
            Location of the FITS file to read. Should probably exist.
        """
        self.hdu_list = pyfits.open(filename)

        try:
            self.fit_mode = self.hdu_list[0].header['PACFMODE']
        except KeyError:
            raise KeyError('Could not determine which fitting recipe was used to produce parameters in {}'.format(filename))

        try:
            self.init_set = self.hdu_list[0].header['INITSET']
        except KeyError:
            raise KeyError('Could not determine which initial value set was used to produce parameters in {}'.format(filename))

        i = 0
        for p in self.required['global']:
            if self.hdu_list[0].header['CUPAR{:03n}'.format(i)] != p:
                raise ValueError('Parameter file is misformed for parameter {}'.format(p))
            i += 1
        for j in range(self.numdrawers):
            for p in self.required['params']:
                if p in self.required['global']:
                    continue
                if self.hdu_list[0].header['CUPAR{:03n}'.format(i)] != '{:}_CS{:02n}'.format(p, j):
                    raise ValueError('Parameter file is misformed for paramter {:}_CS{:02n}'.format(p, j))
                i += 1

        self.CU_params = self.hdu_list[0].data
        self.shape = self.CU_params.shape[:-1]

    def cleanup(self):
        """Remove the temporary file used to hole intermediate results every time __setitem__ was called."""
        os.remove(self.tmp_name)

class TelescopeModelParams():
    """A container for Telescope Model parameters. The primary functions of this object are to write a FITS file that
    contains these parameters in a format that allows Instrument Partners to resample them to whatever field they need
    and to load a set of parameters after they have been resampled.

    The TM parameters are kept in a 3D array of shape (X, Y, L, 6) where (X, Y, L) is the shape of the FOV present in
    the StoCCinG used in fitting.
    """

    def __init__(self, x, y, l, header=None):
        """Initialize the object. All this does is create the empty array that will hold TM parameters.

        Parameters
        ----------
        x : int
            Number of field spatial positions in the x direction

        y : int
            Number of field spatial positions in the y direction

        l : int
            Number of field positions in the l dimension

        header : `astropy.io.fits.header.Header`
            If provided, this object will be initialized with this header
        """
        self.shape = (x, y, l)
        self.tmp_name = 'tmpTM_pars_{}.fits'.format(uuid.uuid4().hex)

        self._init_tmp_file(x, y, l, header=header)

        self.hdu_list = pyfits.open(self.tmp_name, mode='update')
        self.TM_params = self.hdu_list[0].data

    def __del__(self):
        """This is only here to be nice to any developers/testers who forget to call cleanup().
        In general, it is bad practice to rely on this method and cleanup() should be called explicitly.
        """
        try:
            self.cleanup()
        except:
            pass

    def _init_tmp_file(self, x, y, l, header=None):
        """Initialize a temp FITS file that will store the CU Parameters. We do this so that a a single file descriptor
        is used every time __setitem__ is called an the results are flushed to disk.

        Parameters
        ----------
        numparams : int
            The number of fit parameters

        x, y : int
            The shape of the Dresser

        header : `astropy.io.fits.header.Header`
            If provided, initialize the tmp file with this header
        """
        tmp_ph = pyfits.PrimaryHDU(np.zeros((x, y, l, 6)), header=header)
        tmp_ph.header['TMPAR000'] = 'X_12'
        tmp_ph.header['TMPAR001'] = 'tau_12 [rad]'
        tmp_ph.header['TMPAR002'] = 'X_34'
        tmp_ph.header['TMPAR003'] = 'tau_34 [rad]'
        tmp_ph.header['TMPAR004'] = 'X_56'
        tmp_ph.header['TMPAR005'] = 'tau_56 [rad]'

        tmp_ph.writeto(self.tmp_name)

        del tmp_ph.data
        del tmp_ph

    def __repr__(self):
        return "<TelescopeModelParams with shape ({}, {}, {})>".\
            format(*list(self.TM_params.shape))

    def __setitem__(self, key, value):
        """Insert a set of fit parameters from a SoCC or StoCCinG into the TM parameter array.

        Every time an item is set, the current object will be written to disk. This is a somewhat janky way to ensure
        that a crash on looong fits does not result in a total loss of fit results.

        Parameters
        ----------
        key : tuple
            The (x, y, l) spatial location of the SoCC or StoCCinG

        value : `PAC_Pipeline.TelescopeModel.TelescopeModel`
            The best-fit Telescope Model
        """
        xpos, ypos, lpos = key

        if value is None:
            value = SimpleNamespace(**{'x12': 0, 't12': 1,
                                       'x34': 0, 't34': 1,
                                       'x56': 0, 't56': 1})

        self.TM_params[xpos, ypos, lpos, 0] = value.x12
        self.TM_params[xpos, ypos, lpos, 1] = value.t12
        self.TM_params[xpos, ypos, lpos, 2] = value.x34
        self.TM_params[xpos, ypos, lpos, 3] = value.t34
        self.TM_params[xpos, ypos, lpos, 4] = value.x56
        self.TM_params[xpos, ypos, lpos, 5] = value.t56

        self.hdu_list.flush()

    def __getitem__(self, item):
        """Use the currently loaded TM parameters to generate list of TM mirror parameters.

        Parameters
        ----------
        item : tuple
            (x, y, l) location of the CU parameters to load

        Returns
        -------
        list
            List of TM mirror parameters: [x12, t12, x34, t34, x56, t56]
        """
        xpos, ypos, lpos = item

        return self.TM_params[xpos, ypos, lpos, :]

    def writeto(self, filename, header=None, overwrite=True):
        """Write the TM parameters to a FITS file

        The FITS file will have a single HDU with a (X, Y, L, 6). Any header provided will be updated to include the
        parameter names for each entry along the last dimension. If no header is provided a blank one will be created.

        Parameters
        ----------
        filename : str
            Path to save the FITS file

        header : `astropy.io.fits.header.Header`
            An optional header to add to the FITS file. It will be updated with information about the parameter names

        overwrite : bool
            If False then an error will be raised if filename already exists
        """
        if header is not None:
            self.hdu_list[0].header.update(header)

        self.hdu_list[0].writeto(filename, overwrite=overwrite)

    def loadfits(self, filename):
        """Read TM parameters from a FITS file

        The FITS file should have a single HDU with an (X, Y, L, 6) array.

        Parameters
        ----------
        filename : str
            Location of the FITS file to read. Should probably exist.
        """
        self.hdu_list = pyfits.open(filename)
        self.TM_params = self.hdu_list[0].data
        self.shape = self.TM_params.shape[:-1]

    def cleanup(self):
        """Remove the temporary file used to hole intermediate results every time __setitem__ was called."""
        os.remove(self.tmp_name)



#!/usr/bin/python
import os
import pickle
import pkg_resources
import time
import warnings
import asdf
import yaml
from importlib import resources
import numpy as np
from astropy.io import fits as pyfits
import astropy.coordinates as apc
from astropy.time import Time
from dkist_processing_pac import TelescopeModel, CUModel, generic
from dkist_processing_pac.tag import tag
from dkist_processing_pac.M12 import ContinuumPolarization
from dkist_processing_pac.data import S122
from astropy.utils.exceptions import ErfaWarning
from dkist_header_validator import spec122

warnings.filterwarnings('ignore', category=ErfaWarning)
np.random.seed(42)

AZ = np.random.random() * 360.
EL = np.random.random() * 360.
TAB = np.random.random() * 360.

VTF_HEAD = {}

def gen_dummy_ViSP_data(outdir, nummod=8, numsteps=20, shape=(2048, 2048)):

    for m in range(numsteps):

        hdulist = [pyfits.PrimaryHDU()]
        hdulist[0].header['TAZIMUTH'] = AZ
        hdulist[0].header['TELEVATN'] = EL
        hdulist[0].header['TTBLANGL'] = TAB
        hdulist[0].header[S122['ViSP_NumberOfModulatorStates']] = nummod
        hdulist[0].header[S122['PolarizerAngle']] = 360. * (m + 1) / numsteps
        hdulist[0].header[S122['RetarderAngle']] = 360. * (numsteps - m) / numsteps

        for n in range(nummod):
            print(m, n)
            data = np.arange(shape[0] * shape[1], dtype=np.float64).reshape(*shape) * n + m
            ih = pyfits.ImageHDU(data)
            ih.header[S122['ViSP_NumberOfModulatorStates']] = nummod
            ih.header[S122['ViSP_CurrentState']] = n
            hdulist.append(ih)

        outname = '{:}/VISP.C{:03}.FITS'.format(outdir, m)
        pyfits.HDUList(hdulist).writeto(outname, overwrite=True)


def gen_full_SoCC_run(outdir, ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      wave=633., start_time=57530., nummod=8, shape=(1, 10, 10), tab_offset=0,
                      nondata_mask=None, I_sys=55555, df=-500, Q_in=0, U_in=0, V_in=0, use_M12=True,
                      rand_TM=False, spec_122=False, DHS=True, rand_trans=False, SNR=0, muellernoise=0,
                      RN=0, telescope_db=None, seed=42, modmat=None, sin_kwargs=None, allsame=False,
                      dark_signal=0, instrument='visp', suffix='FITS'):

    numsteps = theta_pol_steps.size

    # Assume each step takes 15s
    dt = 15 / (3600 * 24.)
    step_times = np.arange(numsteps) * dt + start_time
    # Within each step each modulator takes 1 second
    dt_m = 1. / (3600 * 24.)

    # This ensures that the true I_sys (the average of the clear measurements) is what the user requested
    I0 = I_sys - 0.5 * (numsteps * df)

    # Load correct instrument-specific stuff
    if instrument.lower() == 'visp':
        from ViSP_Pipeline.utils.gen_fake_data import ViSPDataSet
        from dkist_processing_pac.data import load_visp_keywords
        name_prefix = 'VISP'
        header_func = fill_visp_header
        header_func_args = []
        load_visp_keywords()
        VDS = ViSPDataSet('PolCal', start_time, wave, nummod)
        inst_head = VDS.header()

    elif instrument.lower() == 'dlnirsp':
        from DLNIRSP_Pipeline.utils.gen_fake_data import DLDataSet
        from dkist_processing_pac.data import load_dlnirsp_keywords
        name_prefix = 'DLNIRSP'
        header_func = fill_dlnirsp_header
        header_func_args = []
        load_dlnirsp_keywords()
        DDS = DLDataSet('PolCal', start_time, wave, nummod, 'DLCAM2')
        inst_head = DDS.header()

    elif instrument.lower() in ['cryo-sp', 'cryo-ci']:
        from CryoNIRSP_Pipeline.utils.gen_fake_data import CryoDataSet
        from dkist_processing_pac.data import load_cryo_keywords
        load_cryo_keywords()
        name_prefix = 'CRYONIRSP'
        header_func_args = []
        if instrument.lower() == 'cryo-sp':
            header_func = fill_cryonirsp_sp_header
            CDS = CryoDataSet(1, 'PolCal', 'PolCal4', start_time, wave, nummod)
        else:
            header_func = fill_cryonirsp_ci_header
            CDS = CryoDataSet(2, 'PolCal', 'PolCal4', start_time, wave, nummod)

        inst_head = CDS.header()

    elif instrument.lower() == 'vtf':
        name_prefix = 'VTF'
        header_func = fill_VTF_header
        header_func_args = []
        inst_head = VTF_HEAD


    bool_pol = np.array([i not in ['undefined', 'clear', False] for i in pol_in], dtype=bool)
    bool_ret = np.array([i not in ['undefined', 'clear', False] for i in ret_in], dtype=bool)
    bool_dark = np.array([i in ['Dark', 'DarkShutter'] for i in dark_in], dtype=bool)
    CS = CUModel.CalibrationSequence(theta_pol_steps, theta_ret_steps, bool_pol, bool_ret, bool_dark, step_times,
                                     ret_0_h=ret0_h, dret_h=dret_h,
                                     ret_0_45=ret0_45, dret_45=dret_45,
                                     ret_0_r=ret0_r, dret_r=dret_r,
                                     I_sys=I0, Q_in=Q_in, U_in=U_in, V_in=V_in)
    if rand_trans:
        print('setting random transmission coefficiencts')
        np.random.seed(seed)
        CS.t_ret_0 = np.random.random((1,)) * 0.3 + 0.7
        CS.t_pol_0 = np.random.random((1,)) * 0.3 + 0.7

    CS.set_py_from_database(wave)

    ele = np.zeros(numsteps)
    az = np.zeros(numsteps)
    tab = np.zeros(numsteps)
    for n in range(numsteps):
        obstime = Time(step_times[n], format='mjd')
        tele, taz, ttab = compute_telgeom(obstime)
        ele[n] = tele
        az[n] = taz
        tab[n] = ttab + tab_offset

    TM = TelescopeModel.TelescopeModel(az, ele, tab)

    if telescope_db is None:
        telescope_db = generic.get_default_telescope_db()
    TM.load_from_database(telescope_db, np.mean(step_times), wave)

    if rand_TM:
        np.random.seed(seed)
        TM.x34 += np.random.random() * 0.3 - 0.15
        TM.t34 = (np.random.random() * 60 + 150.) * np.pi / 180
        TM.x56 += np.random.random() * 0.3 - 0.15
        TM.t56 = (np.random.random() * 60 + 150.) * np.pi / 180

    if modmat is None:
        if nummod == 4:
            # Ideal 4-state matrix from DH
            sqrt3 = 1 / np.sqrt(3)
            O = np.array([[ 1.,  sqrt3,  sqrt3, -sqrt3],
                          [ 1.,  sqrt3, -sqrt3,  sqrt3],
                          [ 1., -sqrt3,  sqrt3,  sqrt3],
                          [ 1., -sqrt3, -sqrt3, -sqrt3]], dtype=np.float64)
        elif nummod == 8:
            # Real modulation matrix taken from CBeck's create_synth_cal.pro
            O = np.array([[ 0.693405 ,  0.0833252,  0.598811 , -0.269802 ],
                          [ 0.911968 ,  0.731004 ,  0.327143 , -0.376851 ],
                          [ 0.979729 ,  0.914248 ,  0.084737 ,  0.296836 ],
                          [ 0.711025 ,  0.116364 , -0.224191 ,  0.557482 ],
                          [ 0.603848 , -0.19099  , -0.518671 , -0.0773989],
                          [ 0.862482 ,  0.585429 , -0.38387  , -0.428604 ],
                          [ 1.       ,  0.981222 ,  0.191533 ,  0.0325373],
                          [ 0.789829 ,  0.35732  ,  0.639435 ,  0.233042 ]], dtype=np.float64)
        elif nummod == 10:
            # Modulation matrix for AdW's synthetic ViSP data from mod_matrix_630.out
            O = np.array([[ 1.        ,  0.19155013,  0.80446989, -0.47479524],
                          [ 1.        , -0.65839661,  0.68433984,  0.00466389],
                          [ 1.        , -0.80679413, -0.16112977,  0.48234158],
                          [ 1.        , -0.04856211, -0.56352868,  0.77578117],
                          [ 1.        ,  0.56844858,  0.03324473,  0.77289873],
                          [ 1.        ,  0.19155013,  0.80446989,  0.47479524],
                          [ 1.        , -0.65839661,  0.68433984, -0.00466389],
                          [ 1.        , -0.80679413, -0.16112977, -0.48234158],
                          [ 1.        , -0.04856211, -0.56352868, -0.77578117],
                          [ 1.        ,  0.56844858,  0.03324473, -0.77289873]], dtype=np.float64)
        else:
            raise ValueError('Nummod must be either 8 or 10')
    else:
        if modmat.shape[1] != 4:
            raise ValueError('Provided modulation matrix must have shape (M, 4)')
        O = modmat
        nummod = O.shape[0]

    if muellernoise:
        print('Perturbing Mueller matrices by {} %'.format(muellernoise * 100))
        T = TM.TM + (np.random.random(TM.TM.shape).astype(np.float64) * 2 * muellernoise * np.mean(TM.TM) - muellernoise * np.mean(TM.TM))
        M12 = TM.M12 + (np.random.random(TM.M12.shape).astype(np.float64) * 2 * muellernoise * np.mean(TM.M12) - muellernoise * np.mean(TM.M12))
        CM = CS.CM + (np.random.random(CS.CM.shape).astype(np.float64) * 2 * muellernoise * np.mean(CS.CM) - muellernoise * np.mean(CS.CM))
    else:
        T = TM.TM
        M12 = TM.M12
        CM = CS.CM

    # exp_times contains the obs time for every single modulator state
    mod_delta_t = np.arange(nummod) * dt_m
    exp_times = step_times[:, None] + mod_delta_t[None, :]  # Damn this line is cool
    # exp_times = (N, M)

    S_in = CS.S_in * (1 + df * numsteps / (numsteps - 1) / I0 * np.arange(numsteps))[:, None]
    # (N, 4)

    # This is S_in for every pixel requested (i.e., shape)
    if np.prod(shape) * np.prod(shape) * nummod < 1e5:
        S_in_big = np.dstack([S_in] * nummod * np.prod(shape)).reshape(S_in.shape + (nummod,) + shape)
        mem_err = False
    else:
        print('WARNING! NOT ENOUGH MEMORY TO VARY S_IN FOR EACH PIXEL. FALLING BACK TO SINGLE INPUT FLUX')
        mem_err = True
        OG_shape = shape
        shape = (1,1,1)
        S_in_big = np.dstack([S_in] * nummod).reshape(S_in.shape + (nummod,) + shape)

    S_in_big = np.moveaxis(S_in_big, 2, 1)
    # (N, M, 4, shape)

    if sin_kwargs:
        np.random.seed()
        print('Adding sin perturbation with {}'.format(sin_kwargs))
        sin_pert = sin_perturbation(exp_times, shape, **sin_kwargs)
        # sin_pert = (N, M, shape)
        S_in_big *= 1 + sin_pert[:, :, None, ...]
        # (N, M, 4, shape)

    expander_dims = list(range(-1, -1 * (len(shape) + 1), -1))
    if use_M12:
        print('including M12')
        # T = (N, 4, 4), CM = (N, 4, 4), M12 = (4, 4)
        TCM_exp = np.expand_dims(T @ CM @ M12, [1] + expander_dims)
        # TCM_exp = (N, 1, 4, 4, shape)
    else:
        print('omitting M12')
        TCM_exp = np.expand_dims(T @ CM, [1] + expander_dims)

    # Multiplication basically does TCM @ S_in for all pixels and all CS steps
    S = np.sum(TCM_exp * S_in_big[:, :, None, ...], axis=3)
    # S = (N, M, 4, shape)

    # More array trickery - O_exp is (M, 4, 1, shape)
    O_exp = np.expand_dims(O, [-1 * len(shape) - 1] + expander_dims)
    S_mult = np.moveaxis(S, 0, 2)
    # S_mult = (M 4, N, shape)
    I = np.sum(O_exp * S_mult, axis=1)
    # I = (M, N, shape)

    for n in range(numsteps):

        hdulist = [pyfits.PrimaryHDU()]
        hdulist[0].header.update(inst_head)
        hdulist[0].header['DATE-OBS'] = Time(step_times[n], format='mjd').fits
        hdulist[0].header['DATE-BGN'] = Time(start_time, format='mjd').fits
        hdulist[0].header['DATE-END'] = Time(step_times[-1], format='mjd').fits
        hdulist[0].header[S122['Instrument']] = instrument
        hdulist[0].header[S122['Wavelength']] = wave
        hdulist[0].header['TAZIMUTH'] = az[n]
        hdulist[0].header['TELEVATN'] = ele[n]
        hdulist[0].header['TTBLANGL'] = tab[n]
        hdulist[0].header[S122['InstrumentProgramTask']] = 'polcal'
        hdulist[0].header[S122['GOSPolarizer']] = pol_in[n]
        hdulist[0].header[S122['PolarizerAngle']] = str(theta_pol_steps[n])
        hdulist[0].header[S122['GOSRetarder']] = ret_in[n]
        hdulist[0].header[S122['RetarderAngle']] = str(theta_ret_steps[n])
        hdulist[0].header[S122['GOSLevel0']] = dark_in[n]
        if not DHS:
            hdulist[0].header[S122['ReadNoise']] = RN
        header_func(*([hdulist[0].header, -1, nummod] + header_func_args))

        for m in range(nummod):
            if mem_err:
                data = np.ones(OG_shape, dtype=np.float64) * I[m, n, 0, 0, 0]
            else:
                data = I[m, n]

            if nondata_mask is not None:
                data[nondata_mask] = 0.

            if SNR:
                np.random.seed()
                print('Adding photon noise so SNR = {} at {} counts. <Current counts> = {}'.format(SNR, I_sys,
                                                                                                   np.mean(I[m,n])))
                shotnoise = I_sys / SNR * np.sqrt(np.mean(data) / I_sys)
                print('\tshot noise sigma = {:10.3f}'.format(shotnoise))
                data += np.random.randn(*data.shape) * shotnoise
            if RN:
                np.random.seed()
                print('RN = {:3.1f}'.format(RN))
                data += np.random.randn(*data.shape) * RN

            if dark_signal:
                if type(dark_signal) is tuple and len(dark_signal) == 2:
                    print('Adding dark with mu = {:.2f} and sigma = {:.2f}'.format(*dark_signal))
                    dark = np.random.randn(*data.shape) * dark_signal[1] + dark_signal[0]
                elif type(dark_signal) in [int, float]:
                    print('Adding dark = {:.1f}'.format(dark_signal))
                    dark = dark_signal
                else:
                    raise ValueError('dark_signal must be either a number or a len 2 tuple, you passed {}'.format(dark_signal))

                data += dark

            if allsame:
                ridx = np.unravel_index(np.random.randint(0, data.size), data.shape)
                data = np.ones(shape, dtype=np.float64) * data[ridx]

            if DHS:
                ph = pyfits.PrimaryHDU(np.abs(data).astype(np.uint16), header=hdulist[0].header)
                date = Time(exp_times[n, m], format='mjd')
                ph.header['DATE-OBS'] = date.fits
                header_func(*([ph.header, m, nummod] + header_func_args))
                outname = '{}/{}_{}_{}.{}'.format(outdir, name_prefix, date.datetime.strftime("%Y%m%dT%H%M%S.%f"), m, suffix)
                print(outname)
                ph.writeto(outname, overwrite=True)
            else:
                ih = pyfits.ImageHDU(data, header=hdulist[0].header)
                ih.header['DATE-OBS'] = Time(exp_times[n, m], format='mjd').fits
                header_func(*([ih.header, m, nummod] + header_func_args))
                hdulist.append(ih)

        if not DHS:
            outname = '{:}/{}.C{:03}.{}'.format(outdir, name_prefix, n, suffix)
            pyfits.HDUList(hdulist).writeto(outname, overwrite=True)

    return O, TM, CS, I, az, ele, tab, S_in_big, exp_times


def sin_perturbation(exp_times, shape, amp=0.1, noise=0.1, phase=0, rand_phase=False):

    numsteps, nummod = exp_times.shape

    x = (exp_times.reshape(nummod * numsteps) - exp_times.min()) / (exp_times.max() - exp_times.min()) * 2 * np.pi

    if rand_phase:
        phase_vec = np.random.random(shape) * 2 * np.pi
    else:
        phase_vec = np.ones(shape) * phase
    sin_pert = np.sin(np.expand_dims(x, list(range(-1, -1 * (len(shape) + 1), -1))) + phase_vec[None, ...]) * amp
    if noise > 0:
        sin_pert += np.random.randn(*((numsteps * nummod,) + shape)) * noise * amp

    sin_pert = sin_pert.reshape((numsteps, nummod) + shape)

    return sin_pert

def fill_visp_header(header, m, nummod):
    header[S122['Instrument']] = 'VISP'
    header[S122['ViSP_NumberOfModulatorStates']] = nummod
    header[S122['ViSP_CurrentState']] = m + 1
    header[S122['ViSP_PolarimeterMode']] = 'Full Stokes'
    # header[S122['ViSP_BeamBorder']] = 1080

def fill_dlnirsp_header(header, m, nummod):
    header[S122['DLN_NumberOfModulatorStates']] = nummod
    header[S122['DLN_CurrentState']] = m

def fill_cryonirsp_sp_header(header, m, nummod):
    header[S122['Instrument']] = 'cryo'
    header[S122['CRSP_NumberOfModulatorStates']] = nummod
    header[S122['CRSP_CurrentState']] = m
    header[S122['CRSP_CryoInstrument']] = 1
    header[S122['CRSP_SubtaskName']] = 'Polcal4'
    header[S122['CRSP_PolarimeterMode']] = 'Full Stokes'

def fill_cryonirsp_ci_header(header, m, nummod):
    header[S122['Instrument']] = 'cryo'
    header[S122['CRSP_NumberOfModulatorStates']] = nummod
    header[S122['CRSP_CurrentState']] = m
    header[S122['CRSP_CryoInstrument']] = 2
    header[S122['CRSP_SubtaskName']] = 'Polcal4'
    header[S122['CRSP_PolarimeterMode']] = 'Full Stokes'

def fill_VTF_header(header, m, nummod):
    header[S122['VTF_NumberOfModulatorStates']] = nummod
    header[S122['VTF_CurrentState']] = m

def SoCC_fixed_pol(outputdir, nummod=10, shape=(1, 1, 1), seed=42, itc=True, clear=True, dret=False, rand_S_in=False,
                   rand_Q=False, rand_ret=False, dark=True, wave=633., **full_kwargs):

    polcal_init_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/init_values/polcal_default.asdf')
    with asdf.open(polcal_init_file, 'rb', lazy_load=False, copy_arrays=True) as f:
        polcal_init = f.tree
    wave_idx = np.argmin(np.abs(polcal_init['wave'] - wave))
    params = polcal_init['params']

    if rand_ret:
        np.random.seed(seed)
        ret0_h = np.random.random() * 0.1 + 0.9
        ret0_45 = np.random.random() * 0.1 + 1.4
        ret0_r = np.random.random() * 0.015 + 0.02
    else:
        rng = np.random.default_rng(seed)
        ret0_h = rng.uniform(params['ret0h'][wave_idx, :][0], params['ret0h'][wave_idx, :][2])
        ret0_45 = rng.uniform(params['ret045'][wave_idx, :][0], params['ret045'][wave_idx, :][2])
        ret0_r = rng.uniform(params['ret0r'][wave_idx, :][0], params['ret0r'][wave_idx, :][2])

    if dret:
        # dret from assumption of 0.1 deg per deg C and a ~5 deg C heating over 15 min
        dret_h = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_h += np.random.random() * dret_h * 0.2 - (dret_h * 0.1)
        dret_45 = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_45 += np.random.random() * dret_45 * 0.2 - (dret_45 * 0.1)
        dret_r = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_r += np.random.random() * dret_r * 0.2 - (dret_r * 0.1)
    else:
        dret_h = 0
        dret_45 = 0
        dret_r = 0

    if itc:
        numsteps = 53
        ret_reset = np.array([40., 20.])
    else:
        numsteps = 49
        ret_reset = np.array([40.])

    Q_in = np.random.uniform(params['Q_in'][wave_idx, :][0], params['Q_in'][wave_idx, :][2])
    U_in = 0
    V_in = 0
    if rand_Q:
        Q_in += np.random.random()
    if rand_S_in:
        U_in += np.random.random()
        V_in += np.random.random()

    theta_pol_steps = np.ones(numsteps - 8) * 180.
    theta_ret_steps = np.r_[np.linspace(0, 90, 10), ret_reset]
    for i in range(3):
        theta_ret_steps = np.r_[theta_ret_steps, np.linspace(100, 180, 9) + 90 * i, ret_reset]
    theta_pol_steps = np.r_[np.array([0, 45, 90, 135.]), theta_pol_steps, np.array([180, 225, 270, 315.])]
    theta_ret_steps = np.r_[np.zeros(4), theta_ret_steps, np.zeros(4)]

    pol_in = np.ones(numsteps, dtype=bool)
    ret_in = np.ones(numsteps, dtype=bool)
    ret_in[-4:] = False
    ret_in[:4] = False
    dark_in = np.zeros(numsteps, dtype=bool)

    # Add clears
    if clear:
        pol_in = np.r_[False, pol_in, False]
        ret_in = np.r_[False, ret_in, False]
        dark_in = np.r_[False, dark_in, False]
        theta_pol_steps = np.r_[0., theta_pol_steps, 0.]
        theta_ret_steps = np.r_[0., theta_ret_steps, 0.]

    # Add darks
    if dark:
        pol_in = np.r_[False, pol_in, False]
        ret_in = np.r_[False, ret_in, False]
        dark_in = np.r_[True, dark_in, True]
        theta_pol_steps = np.r_[0., theta_pol_steps, 0.]
        theta_ret_steps = np.r_[0., theta_ret_steps, 0.]

    O, TM, CS, I, _, _, _, _, _ = gen_full_SoCC_run(outputdir, ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                                                    theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                                                    nummod=nummod, shape=shape, Q_in=Q_in, U_in=U_in, V_in=V_in,
                                                    wave=wave, **full_kwargs)

    return ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r, O, TM, CS, I


def SoCC_fixed_ret(outputdir, nummod=10, shape=(1, 1, 1), seed=42, itc=True, clear=True, dret=False, rand_S_in=False,
                   rand_Q=False, rand_ret=False, dark=True, wave=633., **full_kwargs):

    polcal_init_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/init_values/polcal_default.asdf')
    with asdf.open(polcal_init_file, 'rb', lazy_load=False, copy_arrays=True) as f:
        polcal_init = f.tree
    wave_idx = np.argmin(np.abs(polcal_init['wave'] - wave))
    params = polcal_init['params']

    if rand_ret:
        np.random.seed(seed)
        ret0_h = np.random.random() * 0.1 + 0.9
        ret0_45 = np.random.random() * 0.1 + 1.4
        ret0_r = np.random.random() * 0.015 + 0.02
    else:
        rng = np.random.default_rng(seed)
        ret0_h = rng.uniform(params['ret0h'][wave_idx, :][0], params['ret0h'][wave_idx, :][2])
        ret0_45 = rng.uniform(params['ret045'][wave_idx, :][0], params['ret045'][wave_idx, :][2])
        ret0_r = rng.uniform(params['ret0r'][wave_idx, :][0], params['ret0r'][wave_idx, :][2])

    if dret:
        # dret from assumption of 0.1 deg per deg C and a ~5 deg C heating over 15 min
        dret_h = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_h += np.random.random() * dret_h * 0.2 - (dret_h * 0.1)
        dret_45 = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_45 += np.random.random() * dret_45 * 0.2 - (dret_45 * 0.1)
        dret_r = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_r += np.random.random() * dret_r * 0.2 - (dret_r * 0.1)
    else:
        dret_h = 0
        dret_45 = 0
        dret_r = 0

    if itc:
        numsteps = 53
        pol_reset = np.array([0., 40])
        ret_reset = np.array([22.5, 22.5])
    else:
        numsteps = 49
        pol_reset = np.array([0.])
        ret_reset = np.array([22.5])

    Q_in = np.random.uniform(params['Q_in'][wave_idx, :][0], params['Q_in'][wave_idx, :][2])
    U_in = 0
    V_in = 0
    if rand_Q:
        Q_in += np.random.random()
    if rand_S_in:
        U_in += np.random.random()
        V_in += np.random.random()

    theta_pol_steps = np.linspace(0, 360, 37)
    theta_ret_steps = np.ones(37) * 22.5
    theta_ret_steps[1::4] += 5
    theta_ret_steps[2::4] += 5

    for i in range(3):
        idx = 10 + i * (9 + len(pol_reset))
        theta_pol_steps = np.r_[theta_pol_steps[:idx], pol_reset, theta_pol_steps[idx:]]
        theta_ret_steps = np.r_[theta_ret_steps[:idx], ret_reset, theta_ret_steps[idx:]]

    theta_pol_steps = np.r_[theta_pol_steps, pol_reset]
    theta_ret_steps = np.r_[theta_ret_steps, ret_reset]

    theta_pol_steps = np.r_[np.array([0, 45, 90, 135.]), theta_pol_steps, np.array([180, 225, 270, 315.])]
    theta_ret_steps = np.r_[np.zeros(4), theta_ret_steps, np.zeros(4)]

    pol_in = np.ones(numsteps, dtype=bool)
    ret_in = np.ones(numsteps, dtype=bool)
    dark_in = np.zeros(numsteps, dtype=bool)
    ret_in[-4:] = False
    ret_in[:4] = False

    # Add clears
    if clear:
        pol_in = np.r_[False, pol_in, False]
        ret_in = np.r_[False, ret_in, False]
        dark_in = np.r_[False, dark_in, False]
        theta_pol_steps = np.r_[0., theta_pol_steps, 0.]
        theta_ret_steps = np.r_[0., theta_ret_steps, 0.]

    # Add darks
    if dark:
        pol_in = np.r_[False, pol_in, False]
        ret_in = np.r_[False, ret_in, False]
        dark_in = np.r_[True, dark_in, True]
        theta_pol_steps = np.r_[0., theta_pol_steps, 0.]
        theta_ret_steps = np.r_[0., theta_ret_steps, 0.]

    O, TM, CS, I, _, _, _, _, _ = gen_full_SoCC_run(outputdir, ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                                                    theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                                                    nummod=nummod, shape=shape, Q_in=Q_in, U_in=U_in, V_in=V_in,
                                                    wave=wave, **full_kwargs)

    return ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r, O, TM, CS, I

def CS_from_defaults(CS_name):

    if '_CS' not in CS_name:
        CS_name = CS_name + '_CS'
    CS_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/CS_files/{}.txt'.format(CS_name))
    theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in = np.loadtxt(CS_file, unpack=True)

    pol_in = np.array(pol_in, dtype=bool)
    ret_in = np.array(ret_in, dtype=bool)
    dark_in = np.array(dark_in, dtype=bool)

    pol_in = np.array(['Polarizer' if i else 'clear' for i in pol_in])
    ret_in = np.array(['SAR' if i else 'clear' for i in ret_in])
    dark_in = np.array(['DarkShutter' if i else "FieldStop (5arcmin)" for i in dark_in])

    return theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in

def SoCC_from_defaults(outputdir, CS_name, nummod=10, shape=(1, 1, 1), seed=42, dret=False, rand_S_in=False,
                       rand_Q=False, rand_ret=False, wave=633, **full_kwargs):

    polcal_init_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/init_values/polcal_default.asdf')
    with asdf.open(polcal_init_file, 'rb', lazy_load=False, copy_arrays=True) as f:
        polcal_init = f.tree
    wave_idx = np.argmin(np.abs(polcal_init['wave'] - wave))
    params = polcal_init['params']

    if rand_ret:
        np.random.seed(seed)
        ret0_h = np.random.random() * 0.1 + 0.9
        ret0_45 = np.random.random() * 0.1 + 1.4
        ret0_r = np.random.random() * 0.015 + 0.02
    else:
        rng = np.random.default_rng(seed)
        ret0_h = rng.uniform(params['ret0h'][wave_idx, :][0], params['ret0h'][wave_idx, :][2])
        ret0_45 = rng.uniform(params['ret045'][wave_idx, :][0], params['ret045'][wave_idx, :][2])
        ret0_r = rng.uniform(params['ret0r'][wave_idx, :][0], params['ret0r'][wave_idx, :][2])

    if dret:
        # dret from assumption of 0.1 deg per deg C and a ~5 deg C heating over 15 min
        dret_h = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_h += np.random.random() * dret_h * 0.2 - (dret_h * 0.1)
        dret_45 = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_45 += np.random.random() * dret_45 * 0.2 - (dret_45 * 0.1)
        dret_r = 0.5 / 15 * 60 * 24 * np.pi / 180.
        dret_r += np.random.random() * dret_r * 0.2 - (dret_r * 0.1)
    else:
        dret_h = 0
        dret_45 = 0
        dret_r = 0

    Q_in = np.random.uniform(params['Q_in'][wave_idx, :][0], params['Q_in'][wave_idx, :][2])
    U_in = 0
    V_in = 0
    if rand_Q:
        Q_in += np.random.random()
    if rand_S_in:
        U_in += np.random.random()
        V_in += np.random.random()


    theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in = CS_from_defaults(CS_name)

    O, TM, CS, I, _, _, _, _, _ = gen_full_SoCC_run(outputdir, ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                                                    theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                                                    nummod=nummod, shape=shape, Q_in=Q_in, U_in=U_in, V_in=V_in,
                                                    wave=wave, **full_kwargs)

    return ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r, O, TM, CS, I

def SoCC_multi_day(outputdir, numdays=10, nummod=10, shape=(1, 1, 1), rand_TM=True, DHS=True, rand_Q=False, rand_S_in=False,
                   only_pol=False, only_ret=False, spec_122=True, clear=True, df=None, I_sys=None, dret=False, use_M12=True,
                   rand_trans=True, SNR=0, muellernoise=0, seed=42, modmat=None,
                   start_time=57531., CS_name=None, **full_kwargs):

    dt = 0.021 # ~0.5 hours

    ret0_h_list = []
    dret_h_list = []
    ret0_45_list = []
    dret_45_list = []
    ret0_r_list = []
    dret_r_list = []
    t_pol_list = []
    t_ret_list = []
    I_sys_list = []
    Q_list = []
    U_list = []
    V_list = []
    for i in range(numdays):
        print(compute_telgeom(Time(start_time + (1 + dt) * i, format='mjd'))[0])
        outdir = '{}/day{}'.format(outputdir, i)
        os.makedirs(outdir, exist_ok=True)

        if I_sys is None:
            II_sys = np.random.random() * 50000 + 10000
        else:
            II_sys = I_sys
        if df is None:
            ddf = np.random.random() * 200 - 100
        else:
            ddf = df

        if CS_name is not None:
            print('{}: Making Drawer from CS {}'.format(tag(), CS_name))
            o = SoCC_from_defaults(outdir, CS_name, nummod=nummod, shape=shape, DHS=DHS, start_time=start_time + dt * i,
                                   rand_TM=rand_TM, seed=seed + i, spec_122=spec_122, df=ddf,
                                   I_sys=II_sys, dret=dret, rand_Q=rand_Q, use_M12=use_M12,
                                   rand_S_in=rand_S_in, SNR=SNR, muellernoise=muellernoise, rand_trans=rand_trans,
                                   modmat=modmat, **full_kwargs)
        else:
            np.random.seed()
            choice = np.random.random()
            if (choice > 0.5 and not only_ret) or only_pol:
                print('{}: Making Drawer with fixed_pol'.format(tag()))
                o = SoCC_fixed_pol(outdir, nummod=nummod, shape=shape, DHS=DHS, start_time=start_time + dt * i, rand_TM=rand_TM,
                                   seed=seed + i, spec_122=spec_122, ogy=ogy, clear=clear, df=ddf, I_sys=II_sys, dret=dret,
                                   rand_Q=rand_Q, use_M12=use_M12, rand_S_in=rand_S_in,
                                   SNR=SNR, muellernoise=muellernoise, rand_trans=rand_trans, modmat=modmat, **full_kwargs)
            else:
                print('{}: Making Drawer with fixed_ret'.format(tag()))
                o = SoCC_fixed_ret(outdir, nummod=nummod, shape=shape, DHS=DHS, start_time=start_time + dt * i, rand_TM=rand_TM,
                                   seed=seed + i, spec_122=spec_122, ogy=ogy, clear=clear, df=ddf, I_sys=II_sys, dret=dret,
                                   rand_Q=rand_Q, use_M12=use_M12, rand_S_in=rand_S_in,
                                   SNR=SNR, muellernoise=muellernoise, rand_trans=rand_trans, modmat=modmat, **full_kwargs)

        ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r, _, TM, CS, _ = o

        ret0_h_list.append(ret0_h)
        dret_h_list.append(dret_h)
        ret0_45_list.append(ret0_45)
        dret_45_list.append(dret_45)
        ret0_r_list.append(ret0_r)
        dret_r_list.append(dret_r)
        t_pol_list.append(CS.t_pol_0[0])
        t_ret_list.append(CS.t_ret_0[0])
        I_sys_list.append(II_sys)
        Q_list.append(CS.Q_in)
        U_list.append(CS.U_in)
        V_list.append(CS.V_in)

    truth = {'x12': TM.x12,
             't12': TM.t12,
             'x34': TM.x34,
             't34': TM.t34,
             'x56': TM.x56,
             't56': TM.t56,
             'I_sys': I_sys_list,
             't_pol': t_pol_list,
             't_ret': t_ret_list,
             'Q_in': Q_list,
             'U_in': U_list,
             'V_in': V_list,
             'ret_0_h': ret0_h_list,
             'dret_h': dret_h_list,
             'ret_0_45': ret0_45_list,
             'dret_45': dret_45_list,
             'ret_0_r': ret0_r_list,
             'dret_r': dret_r_list}

    with open('{}/truth.pkl'.format(outputdir),'wb') as f:
        pickle.dump(truth, f)
    print_truth('{}/truth.pkl'.format(outputdir))

    return truth


def print_truth(truth_file='truth.pkl'):
    with open(truth_file, 'rb') as f:
        truth = pickle.load(f)

    for k in truth.keys():
        print('{:<10}:'.format(k), end='')
        if type(truth[k]) is list:
            try:
                print(str('{:10.3f}'*len(truth[k])).format(*truth[k]))
            except ValueError:
                print(str('{:10}'*len(truth[k])).format(*truth[k]))
            except Exception:
                print(' CANNOT PRINT')
        else:
            try:
                print('{:10.3f}'.format(truth[k]))
            except ValueError:
                print('{:10}'.format(truth[k]))

def make_M12_map(outdir, simdir='/home/ade/PycharmProjects/ViSP_pipeline/sim_data',
                 low_lim=15, high_lim=75, cont_low=-20, cont_high=None, datanoise=0):

    TM = TelescopeModel.TelescopeModel(0,0,0)
    np.random.seed(13)
    TM.x12 = np.random.random() * 0.35 + 0.9
    TM.t12 = (np.random.random() * 60 + 150.) * np.pi / 180

    for i in range(1000):
        print(i)
        hdus = pyfits.open('{:}/stokes/image.{:04}.fits'.format(simdir,i))
        stokes = np.dstack([h.data for h in hdus[1:]])

        data = np.sum(TM.M12 * stokes[:, :, None, :], axis=3)

        if datanoise:
            data += np.random.randn(*data.shape) * datanoise * np.mean(data)

        hdulist = pyfits.HDUList([pyfits.PrimaryHDU()])
        for j, s in enumerate(['I','Q','U','V']):
            hdulist.append(pyfits.ImageHDU(data[:, low_lim:high_lim, j]))
            hdulist[-1].header['VISP_STK'] = s
            hdulist[-1].header[S122['Wavelength']] = 588
            hdulist[-1].header['DATE-OBS'] = Time(57560, format='mjd').fits

        hdulist.writeto('{}/M12.{:04n}.FITS'.format(outdir, i), overwrite=True)

        I = data[:, cont_low:cont_high, 0]
        Q = data[:, cont_low:cont_high, 1]
        U = data[:, cont_low:cont_high, 2]

        i2q = np.nanmedian(Q / I)
        i2u = np.nanmedian(U / I)

        ContinuumPolarization.update_xtalk_database(i2q, i2u, 'gen_fake_data',
                                                    Time.now().mjd, 666)
        hdus.close()
        del hdus
        del stokes

    big_list = []
    for i in range(1000):
        print(i)
        hdus = pyfits.open('{}/M12.{:04}.FITS'.format(outdir,i))
        big_list.append(np.dstack([h.data for h in hdus[1:]]))
        hdus.close()

    big_map = np.stack(big_list, 1)
    del big_list
    print(big_map.shape)
    print(TM.x12, TM.t12)

    hdulist = pyfits.HDUList([pyfits.PrimaryHDU()])
    for i in range(4):
        hdulist.append(pyfits.ImageHDU(big_map[:, :, :, i]))

    hdulist[0].header[S122['Wavelength']] = 588
    hdulist[0].header['DATE-OBS'] = Time(57560, format='mjd').fits
    hdulist.writeto(outdir + '/full_maps.FITS', overwrite=True)

    return

 #  re-work of Arthur's function by DMH
def compute_telgeom(time_hst):
  dkist_lon = (156 + 15/60.0 + 21.7/3600.0) * (-1)
  dkist_lat = ( 20 + 42/60.0 + 27.0/3600.0)
  hel  = 3040.4
  hloc = apc.EarthLocation.from_geodetic(dkist_lon, dkist_lat, hel)
  sun_body = apc.get_body('sun', time_hst, hloc)  #get the solar ephemeris
  sun_dec = sun_body.dec.value   # Extract declination in degrees
  azel_frame = apc.AltAz(obstime=time_hst, location=hloc)  # Horizon coords
  sun_altaz = sun_body.transform_to(azel_frame)  # Sun in horizon coords
  alt = sun_altaz.alt.value  # Extract altitude
  azi = sun_altaz.az.value   # Extract azimuth

  #  COMPUTE Gregorian image rotation relative to celestial
  #  \cos\theta_{G} = \frac {\sin(Lat) - \sin(El) \sin(Dec)}  {\cos(El)\cos(Dec)}
  r2d = np.double(180) / np.pi
  cos_thetaG_top = np.sin( dkist_lat/r2d ) - np.sin( alt/r2d ) * np.sin(sun_dec/r2d)
  cos_thetaG_bot = np.cos(alt/r2d) * np.cos(sun_dec/r2d)
  thetaG = np.arccos( cos_thetaG_top / cos_thetaG_bot ) * r2d  #Gregorian image rot

  tableang = alt-azi#-thetaG

  return alt, azi, tableang#, thetaG

def convert_cbeck_db(inputfile='/home/ade/PAC/DKIST_programs/telescope_parameters.txt', outputfile='telescope_db.txt'):

    wave, x12, t12, x34, t34, x56, t56 = np.loadtxt(inputfile, usecols=(2, 3, 4, 5, 6, 7, 8), unpack=True, skiprows=1,
                                                    delimiter=',')
    date, time = np.loadtxt(inputfile, usecols=(0, 1), dtype=str, unpack=True, skiprows=1, delimiter=',')

    obstime = np.zeros(wave.size)
    for i in range(obstime.size):
        obstime[i] = Time('{}T{}:00'.format(date[i], time[i]), format='fits').mjd

    stack = np.vstack((obstime, wave, x12, t12 * np.pi / 180, x34, t34 * np.pi / 180, x56, t56 * np.pi / 180)).T

    with open(outputfile, 'w') as f:
        f.write(str('#{:>14}' + '{:>8}' + '{:>7}' * 6 + '\n').format('MJD', 'wave', 'x12', 't12', 'x34', 't34', 'x56',
                                                                     't56'))
        for i in range(obstime.size):
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}' * 6 + '\n').format(*stack[i]))

    return


def convert_cbeck_contrast(inputfile='/home/ade/PAC/polarizer_contrast_py_curve.sav', outputfile=None):
    import scipy.io as spio

    data = spio.readsav(inputfile)

    if outputfile is None:
        outputfile = './dkist_processing_pac/data/py_table.txt'

    N = data['wavel'].size
    with open(outputfile, 'w') as f:
        f.write(
            '# Lookup table for polarizer py values, assuming px = 1\n# File generated on {}\n'.format(time.asctime()))
        f.write('#{:>10}{:>13}\n\n'.format('wave [nm]', 'py'))
        for i in range(N):
            f.write('{:11.1f}{:13.9f}\n'.format(data['wavel'][i], data['py_wavel'][i]))

    return


def fill_spec122_header(header, inst_head):

    DEFAULT_STRING = 'Etaoin Shrdlu'
    DEFAULT_TIME = '1988-05-25T00:00:00'
    DEFAULT_INT = 666
    DEFAULT_FLOAT = 6.283185307179586
    DEFAULT_BOOL = False

    # Sections of SPEC 122 that should be present in VTF data
    sections = ['ao', 'camera', 'dkist-dkist', 'dkist-id', 'fits', 'pac', 'visp', 'wcs', 'wfc', 'ws']

    # Keywords that are automatically generated by the astropy library
    ignore = ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'NAXIS3', 'EXTEND', 'END', 'BLANK']

    for section in sections:
        DC_s122 = yaml.load(resources.read_text(spec122, '{}.yml'.format(section)), Loader=yaml.Loader)
        for key in DC_s122:
            keyword_name = key
            # Check if the keyword should be ignored
            if keyword_name not in ignore:
                keyword_type = DC_s122[key]['type']
                try:
                    comment = DC_s122[key]['comment']
                except KeyError:
                    comment = 'Default comment'
                # Get data of the correct type
                if keyword_type == 'str':
                    if 'DATE' in key:
                        header[keyword_name] = (DEFAULT_TIME, comment)
                    else:
                        header[keyword_name] = (DEFAULT_STRING, comment)
                if keyword_type == 'int':
                    header[keyword_name] = (DEFAULT_INT, comment)
                if keyword_type == 'float':
                    header[keyword_name] = (DEFAULT_FLOAT, comment)
                if keyword_type == 'bool':
                    header[keyword_name] = (DEFAULT_BOOL, comment)

    header.update(inst_head)
    header.update(CAM_HEAD)

    header[S122['InstrumentProgramID']] = 'gen_fake_data'

#TODO: Move this to somewhere in the tests directory
def make_testing_data():
    from ViSP_Pipeline.utils.gen_fake_data import ViSPDataSet
    from dkist_processing_pac.data import load_visp_keywords

    outputdir = pkg_resources.resource_filename('dkist_processing_pac', 'tests/data')
    numsteps = 7
    shape = (3, 2, 1)
    for td in ['run1', 'run2', 'bad_inst', 'bad_date', 'bad_wave', 'bad_mod', 'bad_shape', 'DL', 'cryo', 'lowflux']:
        os.makedirs(outputdir + '/' + td, exist_ok=True)

    # One with fixed ret
    ####################
    ret0_h = np.pi / 2.
    dret_h = 0.5 / 15 * 60 * 24 * np.pi / 180
    ret0_45 = np.pi / 2.
    dret_45 = 0.5 / 15 * 60 * 24 * np.pi / 180
    ret0_r = np.pi / 2.
    dret_r = 0.5 / 15 * 60 * 24 * np.pi / 180

    theta_pol_steps = np.r_[np.array([0]), np.linspace(0, 360, numsteps - 2), np.array([0])]
    theta_ret_steps = np.ones(numsteps) * 25
    theta_ret_steps[::2] -= 2.5
    theta_ret_steps[1::2] += 2.5

    pol_in = np.ones(numsteps, dtype=bool)
    ret_in = np.ones(numsteps, dtype=bool)
    ret_in[-2] = False
    ret_in[2] = False
    dark_in = np.zeros(numsteps, dtype=bool)
    dark_in[-1] = True
    dark_in[0] = True

    pol_in = np.array(['Polarizer' if i else 'clear' for i in pol_in])
    ret_in = np.array(['SAR' if i else 'clear' for i in ret_in])
    dark_in = np.array(['DarkShutter' if i else "FieldStop (5arcmin)" for i in dark_in])

    gen_full_SoCC_run(outputdir + '/run1', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=shape, spec_122=True, wave=620., start_time=57530., DHS=False, RN=9.)

    # DLNIRSP
    #########
    gen_full_SoCC_run(outputdir + '/DL', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=shape, spec_122=True, wave=620., start_time=57530., DHS=False, RN=9.,
                      instrument='dlnirsp')

    # Cryo
    ######
    gen_full_SoCC_run(outputdir + '/cryo', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=shape, spec_122=True, wave=620., start_time=57530., DHS=False, RN=9.,
                      instrument='cryo-sp')

    # No flux
    ######
    theta_pol_steps = np.r_[np.array([0]), np.linspace(0, 360, numsteps - 2), np.array([0])]
    theta_ret_steps = np.ones(numsteps) * 25
    theta_ret_steps[::2] -= 2.5
    theta_ret_steps[1::2] += 2.5

    pol_in = np.ones(numsteps, dtype=bool)
    ret_in = np.ones(numsteps, dtype=bool)
    ret_in[-2] = False
    ret_in[1] = False
    pol_in[-2] = False
    pol_in[1] = False
    dark_in = np.zeros(numsteps, dtype=bool)
    dark_in[-1] = True
    dark_in[0] = True

    pol_in = np.array(['Polarizer' if i else 'clear' for i in pol_in])
    ret_in = np.array(['SAR' if i else 'clear' for i in ret_in])
    dark_in = np.array(['DarkShutter' if i else "FieldStop (5arcmin)" for i in dark_in])

    gen_full_SoCC_run(outputdir + '/lowflux', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=shape, spec_122=True, wave=620., start_time=57530., DHS=False, RN=9.,
                      I_sys=10, df=0)

    # One with fixed pol
    ####################
    ret0_h = np.pi / 2.
    dret_h = 0.5 / 15 * 60 * 24 * np.pi / 180
    ret0_45 = np.pi / 2.
    dret_45 = 0.5 / 15 * 60 * 24 * np.pi / 180
    ret0_r = np.pi / 2.
    dret_r = 0.5 / 15 * 60 * 24 * np.pi / 180

    theta_pol_steps = np.ones(numsteps) * 180
    theta_ret_steps = np.r_[np.array([0]), np.linspace(0, 360, numsteps - 2), np.array([0])]

    pol_in = np.ones(numsteps, dtype=bool)
    ret_in = np.ones(numsteps, dtype=bool)
    ret_in[-2] = False
    ret_in[2] = False
    dark_in = np.zeros(numsteps, dtype=bool)
    dark_in[-1] = True
    dark_in[0] = True

    pol_in = np.array(['Polarizer' if i else 'clear' for i in pol_in])
    ret_in = np.array(['SAR' if i else 'clear' for i in ret_in])
    dark_in = np.array(['DarkShutter' if i else "FieldStop (5arcmin)" for i in dark_in])

    gen_full_SoCC_run(outputdir + '/run2', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=shape, spec_122=True, wave=620., start_time=57560., DHS=False, RN=13.)

    # One with the wrong shape
    ##########################
    gen_full_SoCC_run(outputdir + '/bad_shape', ret0_h, dret_h, ret0_45, dret_45, ret0_r, dret_r,
                      theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in,
                      nummod=10, shape=(10,10), spec_122=True, wave=620., start_time=57560., DHS=False)

    # Data with inconsistent instrument headers
    ###########################################
    load_visp_keywords()
    VDS = ViSPDataSet('PolCal', 57530., 666, 10)
    VISP_HEAD = VDS.header()

    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_inst/VISP1.FITS', overwrite=True)

    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'not_visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_inst/VISP2.FITS', overwrite=True)

    # Data with inconsistent date headers
    ###########################################
    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_date/VISP1.FITS', overwrite=True)

    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header['DATE-BGN'] = 'Um, is this even a date?'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_date/VISP2.FITS', overwrite=True)

    # Data with inconsistent wavelength headers
    ###########################################
    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    primary.header[S122['Wavelength']] = 620
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_wave/VISP1.FITS', overwrite=True)

    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    primary.header[S122['Wavelength']] = 666
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_wave/VISP2.FITS', overwrite=True)

    # Data with inconsistent modulator headers
    ###########################################
    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 10
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_mod/VISP1.FITS', overwrite=True)

    primary = pyfits.PrimaryHDU()
    primary.header.update(VISP_HEAD)
    primary.header[S122['Instrument']] = 'visp'
    primary.header[S122['ViSP_NumberOfModulatorStates']] = 8
    pyfits.HDUList([primary, pyfits.ImageHDU(np.ones((3, 2, 1)), header=primary.header)]).writeto(
            outputdir + '/bad_mod/VISP2.FITS', overwrite=True)


def command_line():
    import argparse

    parser = argparse.ArgumentParser(description='Generate a SoCC that can be used for science data reduction')
    parser.add_argument('output_dir', help='Location to save the data', nargs=1)
    parser.add_argument('-C', '--CS-name', help='Name of pre-defined default Calibration Sequence to use')
    parser.add_argument('-I', '--instrument', help='The instrument to generate PolCal for',
                        choices=['visp', 'dlnirsp', 'cryo-sp', 'cryo-ci', 'vtf'], default='visp')
    parser.add_argument('-m', '--mask', help='File defining non-data pixels. Needed if instrument is "dlnirsp"')
    parser.add_argument('-p', '--fixed-pol',
                        help='Use a CS with a fixed polarizer. The default is to use a fixed retarder',
                        action='store_true')
    parser.add_argument('-N', '--SNR', help='Signal to Noise of data frames (0 = no noise)',
                        nargs=1, default=[0], type=float)
    parser.add_argument('-R', '--read-noise', help='Read noise (0 = no noise)', default=0, type=float)
    parser.add_argument('-M', '--mueller-noise',
                        help='Fractional amount of uniform noise to perturb telescope and CU Mueller matrices (0 - 1)',
                        nargs=1, default=[0], type=float)
    parser.add_argument('-D', '--two-d', help='Make 2D DHS data', action='store_true')
    parser.add_argument('--no-dhs', help='Make CS for direct input to PA&C Modules', action='store_false')

    args = parser.parse_args()

    O540_modmat = np.array([[1, -0.85243,  0.18135, -0.49039],
                            [1, -0.04202, -0.59912, -0.79956],
                            [1,  0.73844,  0.21129, -0.64036],
                            [1, -0.07196,  0.99175, -0.10604],
                            [1, -0.85243,  0.18135,  0.49039],
                            [1, -0.04202, -0.59912,  0.79956],
                            [1,  0.73844,  0.21129,  0.64036],
                            [1, -0.07196,  0.99175,  0.10604]], dtype=np.float32)

    dark_signal = 0
    if args.instrument == 'visp':
        shape = (1, 2160, 2560)
        mask = None
        modmat = None
        nummod = 10
        wave = 588.
        suffix = 'FITS'
        start_time=57531.
        dark_signal = (94., 3.)
    elif args.instrument == 'dlnirsp':
        try:
            mask = pyfits.open(args.mask)[0].data.astype(bool)
        except:
            raise ValueError('Could not load data mask. Make sure you specified a good mask with "-m"')
        nummod = 8
        modmat = O540_modmat
        wave = 633.
        shape = mask.shape
        suffix = 'FITS'
        start_time = 57531.
        dark_signal = (628., 25)
    elif args.instrument == 'cryo-sp':
        shape = (1, 2048, 2048)
        mask = None
        modmat = O540_modmat
        nummod = 8
        wave = 1080.
        suffix = 'FITS'
        start_time=57529.
    elif args.instrument == 'cryo-ci':
        shape = (1, 2048, 2048)
        mask = None
        modmat = O540_modmat
        nummod = 8
        wave = 1080.
        suffix = 'FITS'
        start_time = 57528.
        dark_signal = (3000., 200.)
    elif args.instrument == 'vtf':
        shape = (1, 2048, 2048)
        mask = None
        sqrt3 = 1 / np.sqrt(3)
        modmat = np.array([[ 1.,  sqrt3,  sqrt3,  sqrt3],
                           [ 1.,  sqrt3, -sqrt3, -sqrt3],
                           [ 1., -sqrt3, -sqrt3,  sqrt3],
                           [ 1., -sqrt3,  sqrt3, -sqrt3]], dtype=np.float64)
        nummod = 4
        wave = 633.
        suffix = 'FITS'
        start_time=57527.

    if args.two_d:
        if len(shape) == 3 and shape[0] == 1:
            shape = shape[1:]
        else:
            raise ValueError('OG shape {} is wrong. Weird'.format(shape))
        if args.instrument == 'dlnirsp':
            mask = mask[0, :, :]

    if not args.no_dhs:
        shape = (1, 1, 1)
        mask = None

    output_dir = args.output_dir[0]
    os.makedirs(output_dir, exist_ok=True)
    SoCC_multi_day(output_dir, numdays=1, DHS=args.no_dhs, spec_122=True, rand_TM=False, only_pol=args.fixed_pol, CS_name=args.CS_name,
                   only_ret=True, shape=shape, SNR=args.SNR[0], I_sys=40000, df=-200, nondata_mask=mask, nummod=nummod,
                   muellernoise=args.mueller_noise[0], instrument=args.instrument, wave=wave, dark_signal=dark_signal,
                   modmat=modmat, RN=args.read_noise, suffix=suffix, start_time=start_time)

    os.system('mv {0}/day0/* {0}'.format(output_dir))
    os.removedirs('{}/day0'.format(output_dir))

    print(
        'All done! Add EXACTLY the following lines to the [PolarimetricCalibration] section of your {} config '
        'file:\n'.format(args.instrument))
    print("raw_pol_dir = {}".format(os.path.abspath(output_dir)))


if __name__ == '__main__':
    command_line()



#TODO: Make this a yml with objectification in __init__
elliptical = {'params': ['ret0h', 'dreth',
                         'ret045', 'dret45',
                         'ret0r', 'dretr',
                         't_ret', 't_pol',
                         'I_sys', 'Q_in', 'U_in', 'V_in'],
              'global': ['Q_in', 'U_in', 'V_in']}



import pickle
import time
import numpy as np
import lmfit
from astropy.io import fits as pyfits
from astropy.time import Time
from dkist_processing_pac import TelescopeModel
from dkist_processing_pac.tag import tag
from dkist_processing_pac.data import S122

def main(map_file, telescope_db, outpre, thinx=10, thiny=10, update_db=False, plot=True, **fit_kwargs):
    """Top-level function for fitting t12 from a 3D stokes map

    This function reads in the data, sets up a TelescopeModel and calls the main fitting function. It then processes
    and distributes the results. The main result is a python pickle file containing the fitter output. Optionally,
    the telescope database can be updated and a fit plot generated.

    The 3D stokes map should be a single FITS file with all data in a single ImageHDU. Those data must have shape
    (X, Y, lambda, 4).

    Parameters
    ----------
    map_file : str
        Location of 3D stokes map

    telescope_db : str
        Location of file containing telescope model parameters as function of date and wavelength

    outpre : str
        Prefix to use when saving fit results. The saved fits and, optionally, fit plot will be saved as OUTPRE.pkl
        and OUTPRE.pdf, respectively.

    thinx : int
        Only load every thinx spatial column from the 3D stokes map

    thiny : int
        Only load every thiny spatial row from the 3D stokes map

    update_db : bool
        If True, update the telescope_db with the new t12 value

    plot : bool
        If True, save a plot showing the Q/V correlation coefficient as a function of t12

    fit_kwargs : dict
        Any arguments to be passed to lmfit.minimizer.Minimizer.minimize().

    Returns
    -------
    `lmfit.minimizer.MinimizerResult`
        Fit result object for t12 fit

    `TelescopeModel.TelescopeModel`
        Telescope Model with best-fit parameters
    """
    hdus = pyfits.open(map_file)
    wavelength = hdus[0].header[S122['Wavelength']]
    try:
        obstime = Time(hdus[0].header['DATE-OBS'], format='fits')
    except ValueError:
        obstime = Time(hdus[0].header['DATE-OBS'], format='iso')

    print('{}: loaded map file {}'.format(tag(), map_file))

    full_map = np.stack([h.data[::thiny, ::thinx, :] for h in hdus[1:]], -1)
    hdus.close()
    del hdus
    print('{}: extracted map with every ({}, {}) other pixel'.format(tag(), thinx, thiny))

    print('{}: initializing Telescope Model'.format(tag()))
    TM = TelescopeModel.TelescopeModel(0,0,0)
    TM.load_from_database(telescope_db, obstime.mjd, wavelength)

    print('{}: starting t12 fitter'.format(tag()))
    # I love that this if statement works in all cases
    if ('method' not in fit_kwargs.keys() or fit_kwargs['method'] == 'brute') and plot:
        fit_kwargs['keep'] = 'all'
    fit_out = fit_t12(full_map, TM, **fit_kwargs)

    if update_db:
        print('{}: updating {} with new t12 value'.format(tag(), telescope_db))
        TM.save_to_database(telescope_db, obstime.mjd, wavelength)

    if plot:
        print('{}: generating fit plot'.format(tag()))
        plotfile = outpre + '.pdf'
        plot_fit(fit_out, plotfile)

    with open(outpre + '.pkl', 'wb') as f:
        print('{}: saving results to {}'.format(tag(), f.name))
        pickle.dump(fit_out, f)

    return fit_out, TM

def compute_ratio(QUV, I):
    """Compute normalized Stokes parameter.

    See eq. 1 in "DKIST Polarization Calibration: Determination of M12 group parameters", C. Beck Feb. 2018
    """
    return np.nansum(np.abs(QUV) / I, axis=2)

def correlation_coefficient(QU, V):
    """Compute the linear correlation coefficient between two Stokes signals.

    See eq. 2 in "DKIST Polarization Calibration: Determination of M12 group parameters", C. Beck Feb. 2018

    To minimize the impact of noisy data we use medians instead of sums. The logic being that using a mean is
    mathematically equivalent to sum (because it's a ratio) and a median is more robust to outliers than a mean.
    """
    V_term = V - np.median(V)
    QU_term = QU - np.median(QU)

    numerator = np.median(V_term * QU_term)
    denominator = np.sqrt(np.median(V_term**2)) * np.sqrt(np.median(QU_term**2))

    return numerator / denominator

def total_correlation(params, full_maps, TM, t1):
    """The minimization function for fitting t12

    First we update the TelescopeModel, then we recompute the 3D stokes map with M12 removed, and finally compute the
    linear correlation coefficient.

    Currently only the correlation between stokes Q and V is considered.

    Parameters
    ----------
    params : `lmfit.parameter.Parameters`
        The parameters to be fit. Should just be t12

    full_maps : numpy.ndarray
        3D stokes maps

    TM : `TelescopeModel.TelescopeModel`
        Model describing the state of DKIST's mirrors. We only use the .M12 property from this object.

    t1 : float
        Time at which the fit started. Used for status updates.

    Returns
    -------
    float
        The linear correlation coefficient between Q and V
    """
    parvals = params.valuesdict()
    TM.t12 = parvals['t12']

    new_map = np.sum(np.linalg.inv(TM.M12) * full_maps[:, :, :, None, :], axis=4)

    Q = compute_ratio(new_map[:, :, :, 1], new_map[:, :, :, 0])
    U = compute_ratio(new_map[:, :, :, 2], new_map[:, :, :, 0])
    V = compute_ratio(new_map[:, :, :, 3], new_map[:, :, :, 0])

    C_q = correlation_coefficient(Q, V)
    C_u = correlation_coefficient(U, V)
    #C_tot = np.sqrt(C_q**2 + C_u**2)

    print_status(TM, C_u, t1)

    return C_u

def fit_t12(full_maps, TM, method='brute', **fit_kwargs):
    """The main fitting function for t12

    First we initialize the fit parameters and then send them off to lmfit for fitting.

    Parameters
    ----------
    full_maps : numpy.ndarray
        The 3D stokes maps

    TM : `TelescopeModel.TelescopeModel`
        Model describing the state of DKIST's mirrors. We only use the .M12 property from this object.

    method : str
        Fitting method to use. Any method available to `lmfit` is valid, but this should probably be 'brute'

    fit_kwargs : dict
        Any arguments to pass to `lmfit.minimize`

    Returns
    -------
    `lmfit.minimize.MinimizerResult`
        Fit result object for t12 fit
    """
    if method == 'brute' and 'Ns' not in fit_kwargs.keys():
        fit_kwargs['Ns'] = 100

    params = lmfit.Parameters()
    params.add('t12', value=TM.t12, min=145 * np.pi / 180., max=215 * np.pi / 180.)

    print('{}: starting minimizer with method {} and kwargs {}'.format(tag(), method, fit_kwargs))
    print_header()

    t1 = time.time()
    mini = lmfit.Minimizer(total_correlation, params, fcn_args=(full_maps, TM, t1))
    fit_out = mini.minimize(method=method, params=params, **fit_kwargs)

    print('\n{:}: minimization completed. Best t12 value: {:7.4f}'.format(tag(), fit_out.params['t12'].value))

    return fit_out

def print_header():
    """Print the header that defines columns in the fit status message
    """
    header_str = '{:>20}{:>20}{:>20}'.format('t12', 'C', 't_elapsed')

    print(header_str)

def print_status(TM, C_tot, t1):
    """Print current parameter values, the value of the linear correlation coefficient, and the elapsed fit time
    """
    status_str = '{:}{:20.6e}{:20.6e}'.format('\r',TM.t12,C_tot)

    dt = time.time() - t1
    time_unit = ' s'
    if dt > 120:
        dt /= 60
        time_unit = ' m'
    status_str += '{:18.1f}{:}'.format(dt, time_unit)

    print(status_str, end="", flush=True)

def plot_fit(fit_obj, output):
    """Plot the linear correlation coefficient as a function of t12

    This function is only valid if the 'brute' method was used to fit t12.

    Parameters
    ----------
    fit_obj : `lmfit.minimize.MinimizerResult`
        Fit result object for t12 fit

    output : str
        Location to save the plot
    """
    import matplotlib.pyplot as plt

    if fit_obj.method != 'brute':
        print("{}: Fitting method was not 'brute'. Cannot generate plot.".format(tag()))
        return

    a = np.array([[c.params['t12'].value, c.score] for c in fit_obj.candidates]).T

    ax = plt.figure().add_subplot(111)
    ax.set_xlabel(r'$\tau_{12}$')
    ax.set_ylabel(r'$C_{QV}$')

    ax.plot(a[0], a[1], '.', color='k', ms=1)
    bestval = fit_obj.params['t12'].value
    ax.axvline(bestval, ls=':', color='r', lw=0.8)
    ymax = ax.get_ylim()[1]
    ax.text(bestval, ymax, r'$\tau_{12} = $' + '{:5.3f}'.format(bestval),
            horizontalalignment='center', verticalalignment='bottom')

    ax.figure.savefig(output)

    return

def command_line():
    import argparse
    import os
    import sys

    parser = argparse.ArgumentParser(description="Compute the M12 tau parameter using the baseline method. This method "
                                                 "requires a single file containing 3D Stokes maps of a region with "
                                                 "strong polarization. The dimensions of this file should be "
                                                 "(X, Y, lambda, 4).",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('map_file', help='FITS file containing 3D stokes maps')
    parser.add_argument('outpre', help='Output prefix to save fitting results. Results will live in OUTPRE.pkl')
    parser.add_argument('-t', '--telescope-db', help='Database listing telescope parameters as function of obstime and '
                                                     'wavelength', nargs=1, default=['telescope_db.txt'])
    parser.add_argument('-x', '--thin-x', help='Only load 1 in every thin-x x dimension spatial locations',
                        nargs=1, default=[10], type=int)
    parser.add_argument('-y', '--thin-y', help='Only load in every thin-y y dimension spatial locations',
                        nargs=1, default=[10], type=int)
    parser.add_argument('-u', '--update-db', help='Update the telescope database with the new tau_12 value',
                        action='store_true')
    parser.add_argument('-p', '--plot', help='Generate plot of tau_12 fit. Plot will be saved to OUTPRE.pdf',
                        action='store_true')

    args = parser.parse_args()

    if not os.path.exists(args.map_file):
        print('Could not find map file {}. Aborting'.format(args.map_file))
        sys.exit(1)

    if not os.path.exists(args.telescope_db[0]):
        print('Could not find telescope db {}. Aborting'.format(args.telescope_db[0]))
        sys.exit(1)

    main(args.map_file, args.telescope_db[0], args.outpre,
         update_db=args.update_db, plot=args.plot, thinx=args.thin_x[0], thiny=args.thin_y[0])
